import os
from pathlib import Path

from unpage.config import manager


def get_pid_file() -> Path:
    """Get the PID file path for the active profile graph build"""
    return manager.get_active_profile_directory() / "graph_build.pid"


def get_log_file() -> Path:
    """Get the log file path for the active profile graph build"""
    return manager.get_active_profile_directory() / "graph_build.log"


def is_process_running(pid: int) -> bool:
    """Check if a process with given PID is running"""
    try:
        # Send signal 0 to check if process exists (doesn't actually kill it)
        os.kill(pid, 0)
        return True
    except (OSError, ProcessLookupError):
        return False


def check_and_create_lock() -> bool:
    """Returns True if we can proceed, False if already running"""
    pid_file = get_pid_file()

    if pid_file.exists():
        try:
            existing_pid = int(pid_file.read_text().strip())
            if is_process_running(existing_pid):
                return False
            else:
                # Stale PID file, remove it
                pid_file.unlink()
        except (ValueError, FileNotFoundError):
            # Corrupted or missing file, remove it
            pid_file.unlink(missing_ok=True)

    return True


def create_pid_file(pid: int) -> None:
    """Create PID file with given process ID"""
    pid_file = get_pid_file()
    pid_file.parent.mkdir(parents=True, exist_ok=True)
    pid_file.write_text(str(pid))


def cleanup_pid_file() -> None:
    get_pid_file().unlink(missing_ok=True)



from unpage.knowledge import Node


class RootlyIncident(Node):
    """A Rootly incident node."""

    @property
    def incident_id(self) -> str:
        """Get the Rootly incident ID."""
        return self.raw_data["id"]

    @property
    def title(self) -> str:
        """Get the incident title."""
        return self.raw_data.get("attributes", {}).get("title", "")

    @property
    def status(self) -> str:
        """Get the incident status."""
        return self.raw_data.get("attributes", {}).get("status", "")

    async def get_identifiers(self) -> list[str | None]:
        """Get identifiers for this incident."""
        return [
            *await super().get_identifiers(),
            self.incident_id,
            self.title,
        ]



from unpage.plugins.datadog.nodes.base import DatadogEntityNode


class DatadogApi(DatadogEntityNode):
    """A node for a Datadog API."""

    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["id"],
            f"api:default/{self.raw_data['attributes']['name']}",
        ]



from unpage.plugins.aptible.nodes.base import AptibleNode


class AptibleAwsInstance(AptibleNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["name"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            (self.raw_data["instance_id"], "represents"),
        ]



import os
from pathlib import Path

from pydantic import BaseModel, Field
from pydantic_yaml import parse_yaml_file_as, parse_yaml_raw_as

from unpage.agent.analysis import Agent
from unpage.config import manager
from unpage.warnings import filter_all_warnings

TEMPLATE_DIR = Path(__file__).parent / "templates"


filter_all_warnings()


if os.environ.get("MLFLOW_TRACKING_URI"):
    import logging

    import mlflow

    mlflow_logger = logging.getLogger("mlflow")
    mlflow_logger.setLevel(logging.WARNING)
    print("> unpage: Found MLFLOW_TRACKING_URI environment variable, running `mlflow.autolog()`")
    mlflow.autolog()
    mlflow.set_experiment("Unpage Agent")


def get_agent_templates() -> list[str]:
    """Get the contents of the agent template for the given agent name."""
    return [
        str(template_file.relative_to(TEMPLATE_DIR).with_suffix(""))
        for template_file in TEMPLATE_DIR.glob("**/*.yaml")
    ]


def get_agent_template(agent_name: str) -> str:
    """Get the contents of the agent template for the given agent name."""
    return (Path(__file__).parent / "templates" / f"{agent_name}.yaml").read_text()


class AgentWithDescriptionOnly(BaseModel):
    description: str = Field(description="A description of the agent and when it should be used")


def get_agent_template_description(agent_name: str) -> str:
    text = get_agent_template(agent_name)
    yml = parse_yaml_raw_as(AgentWithDescriptionOnly, text)
    return yml.description.strip()


def get_agents() -> list[str]:
    """Get the names of the agents in the active profile's config directory."""
    agents_dir = manager.get_active_profile_directory() / "agents"
    agents = [
        str(agent_file.relative_to(agents_dir).with_suffix(""))
        for agent_file in agents_dir.glob("**/*.yaml")
    ]
    # Always include default agent
    if "default" not in agents:
        agents.append("default")
    return agents


def get_agent_file(agent_name: str) -> Path:
    return manager.get_active_profile_directory() / "agents" / f"{agent_name}.yaml"


def load_agent(agent_name: str) -> Agent:
    """Load the agent with the given name from the active profile."""
    agent_file = get_agent_file(agent_name)

    # If they're trying to load the default agent and it doesn't exist, create it.
    if not agent_file.exists() and agent_name == "default":
        agent_file.parent.mkdir(parents=True, exist_ok=True)
        agent_file.touch()
        agent_file.write_text(get_agent_template(agent_name))

    agent = parse_yaml_file_as(Agent, agent_file)
    if not agent.name:
        agent.name = agent_file.stem
    return agent


def delete_agent(agent_name: str) -> None:
    """Delete the agent with the given name from the active profile."""
    agent_file = manager.get_active_profile_directory() / "agents" / f"{agent_name}.yaml"
    agent_file.unlink()



import contextlib
import json
import sys
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, cast

import questionary
import rich
from pydantic import BaseModel
from questionary import Choice
from rich.console import Console
from rich.panel import Panel

from unpage.agent.analysis import Agent, AnalysisAgent
from unpage.agent.utils import (
    get_agent_file,
    get_agent_template_description,
    get_agent_templates,
    load_agent,
)
from unpage.cli.agent._app import agent_app
from unpage.cli.agent.create import create_agent
from unpage.cli.configure import welcome_to_unpage
from unpage.config import Config, PluginConfig, PluginSettings, manager
from unpage.plugins.base import REGISTRY, PluginManager
from unpage.plugins.pagerduty.plugin import PagerDutyPlugin
from unpage.plugins.rootly.plugin import RootlyPlugin
from unpage.telemetry import UNPAGE_TELEMETRY_DISABLED, CommandEvents, hash_value
from unpage.utils import confirm, edit_file, select

if TYPE_CHECKING:
    from unpage.plugins.pagerduty.models import PagerDutyIncident
    from unpage.plugins.rootly.models import RootlyIncident


def _panel(text: str) -> None:
    console = Console()
    console.print(Panel(f"[bold]{text}[/bold]", width=80))


@agent_app.command
async def quickstart() -> None:
    """Get up-and-running with an incident agent in less than 5 minutes!"""
    e = CommandEvents("agent quickstart")
    await e.send("start")
    welcome_to_unpage()
    if not await _quickstart_intro():
        await e.send("abandon", {"abandon_reason": "user_declined_intro"})
        return
    rich.print("")
    agent = await _select_template_and_create_agent(e)
    agent = await _edit_agent(agent, e)
    required_plugins, required_plugin_names_that_need_config = await _plugins_to_config_for_agent(
        agent, e
    )
    while not await confirm("Ready to configure these plugins? (No to edit the agent file again)"):
        agent = await _edit_agent(agent, e)
        (
            required_plugins,
            required_plugin_names_that_need_config,
        ) = await _plugins_to_config_for_agent(agent, e)
    rich.print("")
    rich.print("")
    cfg = await _configure_plugins(
        agent, required_plugins, required_plugin_names_that_need_config, e
    )
    plugin_manager = PluginManager(cfg)
    await _demo_an_incident(agent, plugin_manager, e)
    await e.send("complete")


async def _quickstart_intro() -> bool:
    rich.print("""
Unpage is the open source framework for building SRE agents with infrastructure context and secure access to any dev tool.

This quickstart flow will show you how easily you can build your own custom agents for automation. Here's what it will entail:

• Create your first agent. Choose from our pre-defined templates, or build your own from scratch!
• Configure the agent. Give it access to the tools & context it needs
• Run the agent with a test payload to assess the output
""")
    return await confirm("That's it! Ready to get started?")


async def _select_template_and_create_agent(e: CommandEvents) -> Agent:
    _panel("Create your first agent")
    rich.print(
        "First, which agent would you like to try? Choose a template, or make one from scratch. If it's your first time, we recommend starting with a template."
    )
    rich.print("")
    rich.print("----------")
    rich.print("")
    demo_template_names = [
        "default",
        *sorted(t for t in get_agent_templates() if t != "default" and t != "blank"),
    ]
    choices = [
        Choice(
            title=t,
            value=t,
            description=get_agent_template_description(t),
        )
        for t in demo_template_names
    ]
    choices.append(
        Choice(
            title="Build my own from scratch",
            value="blank",
            description="Build your own agent from scratch",
        )
    )
    template_selected = await select(
        "Select a demo agent:",
        choices=choices,
    )
    rich.print("")
    agent_name = f"demo_quickstart__{template_selected}"
    await create_agent(
        agent_name=agent_name,
        overwrite=True,
        template=template_selected,
    )
    await e.send(
        "agent created", {"template": template_selected, "agent_name_hash": hash_value(agent_name)}
    )
    rich.print("")
    rich.print(
        f"Great! You selected {template_selected if template_selected != 'blank' else 'to build your own template'}. When you're ready, we'll open the agent's configuration file in your default editor so you can (optionally) make changes. Make note of the tools that the agent has access to, as this will determine the plugins we'll need to setup before we can test the agent."
    )
    rich.print("")
    return load_agent(agent_name)


async def _edit_agent(agent: Agent, e: CommandEvents) -> Agent:
    _panel("Edit your agent")
    await questionary.press_any_key_to_continue("Hit [enter] to open the editor").unsafe_ask_async()
    agent_file = get_agent_file(agent.name)
    orig_hash = hash_value(agent_file.read_text())
    await edit_file(agent_file)
    await e.send(
        "agent edited",
        {
            "agent_name_hash": hash_value(agent.name),
            "agent_file_changed": hash_value(agent_file.read_text()) != orig_hash,
        },
    )
    rich.print("")
    rich.print(f"You successfully edited the {agent.name} agent! ✨")
    rich.print("")
    return load_agent(agent.name)


async def _interactive_plugin_config(
    plugin_name: str, existing_plugin_settings: PluginSettings | None
) -> PluginSettings:
    plugin_cls = REGISTRY[plugin_name]
    plugin = plugin_cls(
        **{
            **plugin_cls.default_plugin_settings,
            **(existing_plugin_settings if existing_plugin_settings else {}),
        }
    )
    return await plugin.interactive_configure()


async def _plugin_settings_valid(plugin_name: str, plugin_settings: PluginSettings) -> bool:
    plugin_cls = REGISTRY[plugin_name]
    plugin = plugin_cls(**plugin_settings)
    rich.print(f"Validating {plugin.name}...")
    try:
        await plugin.validate_plugin_config()
    except Exception as ex:
        rich.print(f"Error validating {plugin.name}:\n{ex}")
        return False
    rich.print(f"[green]{plugin.name} configuration is valid![/green]")
    return True


async def _plugins_to_config_for_agent(
    agent: Agent, e: CommandEvents
) -> tuple[list[str], list[str]]:
    _panel("Configure the agent")
    rich.print(
        "Before we test the agent, we need to configure some plugins. Based on the tools this agent has access to, it looks like we'll need API keys for the following:"
    )
    rich.print("")
    required_plugin_names = [
        "llm",
        *sorted(a for a in agent.required_plugins_from_tools() if a != "llm"),
    ]
    required_plugin_names_that_need_config = [
        plugin_name
        for plugin_name in required_plugin_names
        if "interactive_configure" in REGISTRY[plugin_name].__dict__
        and callable(REGISTRY[plugin_name].interactive_configure)
    ]
    await e.send(
        "plugins identified",
        {
            "required_plugins": required_plugin_names,
            "required_plugins_that_need_config": required_plugin_names_that_need_config,
        },
    )
    for plugin_name in required_plugin_names_that_need_config:
        rich.print(f"• {plugin_name.upper() if plugin_name == 'llm' else plugin_name.capitalize()}")
    rich.print("")
    rich.print(
        "Don't worry, the LLM won't see your API keys. They're only used in the configuration file to make the tool calls work."
    )
    rich.print("")
    rich.print(
        "Don't use one of these services? You can still use this agent! Learn more in our Getting Started guide at: https://docs.unpage.ai/?utm_source=unpage"
    )
    rich.print("")
    return required_plugin_names, required_plugin_names_that_need_config


async def _configure_plugins(
    agent: Agent,
    required_plugin_names: list[str],
    required_plugin_names_that_need_config: list[str],
    e: CommandEvents,
) -> Config:
    existing_plugins: dict[str, PluginConfig] = {}
    try:
        existing_config = manager.get_active_profile_config()
        existing_plugins = existing_config.plugins
    except Exception:  # noqa: S110 try-except-pass
        pass
    plugins = {}
    for i, plugin_name in enumerate(required_plugin_names_that_need_config):
        existing_plugin_settings = None
        if plugin_name in existing_plugins:
            existing_plugin_settings = existing_plugins[plugin_name].settings
        step_number = i + 1
        rich.print(
            f"[bold] {step_number}. {plugin_name.upper() if plugin_name == 'llm' else plugin_name.capitalize()} configuration[/bold]"
        )
        rich.print("-" * 80)
        attempts = 1
        plugin_config_validated = False
        while True:
            plugin_settings = await _interactive_plugin_config(
                plugin_name=plugin_name,
                existing_plugin_settings=existing_plugin_settings,
            )
            if plugin_config_validated := await _plugin_settings_valid(
                plugin_name, plugin_settings
            ):
                plugins[plugin_name] = PluginConfig(enabled=True, settings=plugin_settings)
                rich.print("")
                break
            rich.print(f"Validation failed for {plugin_name}")
            if not await confirm("Retry?"):
                rich.print(f"⏭️ Skipping configuration of {plugin_name}")
                break
            rich.print("")
            attempts += 1
        await e.send(
            f"plugin configured {plugin_name}",
            {"attempts": attempts, "plugin_config_validated": plugin_config_validated},
        )
        rich.print("")
    cfg = manager.get_empty_config(
        profile=manager.get_active_profile(),
        telemetry_enabled=not UNPAGE_TELEMETRY_DISABLED,
        plugins={
            **plugins,
            **{
                plugin_name: PluginConfig(
                    enabled=True,
                    settings=REGISTRY[plugin_name].default_plugin_settings,
                )
                for plugin_name in required_plugin_names
                if plugin_name not in required_plugin_names_that_need_config
            },
        },
    )
    cfg.save()
    rich.print("")
    return cfg


async def _provide_incident_id_or_url(pd: PagerDutyPlugin) -> str | None:
    """Prompt user to enter a PagerDuty incident ID or URL."""
    while True:
        answer = await questionary.text(
            "Enter PagerDuty incident ID or URL:",
        ).unsafe_ask_async()

        if not answer:
            return None

        incident_id = answer
        if "/" in answer:
            incident_id = [x for x in answer.split("/") if x][-1]

        try:
            incident = await pd.get_incident_by_id(incident_id)
            return incident.model_dump_json(indent=2)
        except Exception as ex:
            rich.print(f"[red]Failed to retrieve incident with id {incident_id}: {ex}[/red]")
            if not await confirm("Retry with another ID or URL?"):
                return None


async def _select_from_recent_incidents(
    pd: PagerDutyPlugin, incidents_to_consider: int = 20
) -> str | None:
    incidents: list[PagerDutyIncident] = []
    console = Console()

    with console.status("Fetching recent incidents...", spinner="dots") as status:
        async for incident in pd.recent_incident_payloads():
            incidents.append(incident.incident)
            if len(incidents) >= incidents_to_consider:
                break
        status.update("Done 🎉")

    if not incidents:
        rich.print("[yellow]No recent incidents found.[/yellow]")
        return None

    enable_search = len(incidents) > 10
    incident_id = await select(
        "Select an incident:",
        choices=[
            Choice(
                f"{i.title[:60]}... [{i.urgency}]"
                if len(i.title) > 60
                else f"{i.title} [{i.urgency}]",
                value=i.id,
            )
            for i in incidents
        ],
        use_search_filter=enable_search,
        use_jk_keys=not enable_search,
    )

    for incident in incidents:
        if incident.id == incident_id:
            return incident.model_dump_json(indent=2)

    return None


async def _provide_rootly_incident_id_or_url(rootly: RootlyPlugin) -> str | None:
    """Prompt user to enter a Rootly incident ID or URL."""
    while True:
        answer = await questionary.text(
            "Enter Rootly incident ID or URL:",
        ).unsafe_ask_async()

        if not answer:
            return None

        incident_id = answer
        if "/" in answer:
            incident_id = [x for x in answer.split("/") if x][-1]

        try:
            incident = await rootly.get_incident_by_id(incident_id)
            return incident.model_dump_json(indent=2)
        except Exception as ex:
            rich.print(f"[red]Failed to retrieve incident with id {incident_id}: {ex}[/red]")
            if not await confirm("Retry with another ID or URL?"):
                return None


async def _select_from_recent_rootly_incidents(
    rootly: RootlyPlugin, incidents_to_consider: int = 20
) -> str | None:
    incidents: list[RootlyIncident] = []
    console = Console()

    with console.status("Fetching recent Rootly incidents...", spinner="dots") as status:
        async for incident_payload in rootly.recent_incident_payloads():
            incidents.append(incident_payload.incident)
            if len(incidents) >= incidents_to_consider:
                break
        status.update("Done 🎉")

    if not incidents:
        rich.print("[yellow]No recent Rootly incidents found.[/yellow]")
        return None

    enable_search = len(incidents) > 10
    incident_id = await select(
        "Select a Rootly incident:",
        choices=[
            Choice(
                f"{incident.attributes.get('title', 'Unknown Title')[:60]}... [{incident.attributes.get('status', 'Unknown')}]"
                if len(incident.attributes.get("title", "")) > 60
                else f"{incident.attributes.get('title', 'Unknown Title')} [{incident.attributes.get('status', 'Unknown')}]",
                value=incident.id,
            )
            for incident in incidents
        ],
        use_search_filter=enable_search,
        use_jk_keys=not enable_search,
    )

    for incident in incidents:
        if incident.id == incident_id:
            return incident.model_dump_json(indent=2)

    return None


async def _use_test_payload_from_agent(agent: Agent) -> str | None:
    """Use a test payload from the agent definition."""
    if not agent.test_payloads:
        return None

    if len(agent.test_payloads) == 1:
        return next(iter(agent.test_payloads.values())).payload

    choices = [
        Choice(test_name, value=payload) for test_name, payload in agent.test_payloads.items()
    ]

    selected_test = await select(
        "Select a test payload:",
        choices=choices,
    )

    return agent.test_payloads[selected_test].payload


async def _provide_json_directly() -> str | None:
    """Prompt user to enter JSON payload directly."""
    rich.print("Enter the JSON payload (press Ctrl+D when done):")
    rich.print("[dim]Tip: You can paste multi-line JSON here[/dim]")

    lines = []
    try:
        while True:
            line = await questionary.text("").unsafe_ask_async()
            if line is None:
                break
            lines.append(line)
    except (EOFError, KeyboardInterrupt):
        pass

    payload_text = "\n".join(lines)

    if not payload_text.strip():
        rich.print("[yellow]No payload provided.[/yellow]")
        return None

    try:
        parsed = json.loads(payload_text)
        return json.dumps(parsed, indent=2)
    except Exception:
        return payload_text


async def _demo_an_incident(agent: Agent, plugin_manager: PluginManager, e: CommandEvents) -> None:
    _panel("Test out your new agent!")
    rich.print(f"You're ready to test the new {agent.name} agent!")
    rich.print(
        "There are many ways to provide an incident for testing. Use the arrows to confirm your preference:"
    )
    rich.print("")

    class PayloadOption(BaseModel):
        title: str
        func: Callable[[], Awaitable[str | None]]

    options = []
    if plugin_manager.config_has_plugin("pagerduty"):
        pd = cast("PagerDutyPlugin", plugin_manager.get_plugin("pagerduty"))
        options.extend(
            [
                PayloadOption(
                    title="Provide a PagerDuty incident ID or URL",
                    func=lambda: _provide_incident_id_or_url(pd),
                ),
                PayloadOption(
                    title="Select from a list of 20 most recent PagerDuty incidents",
                    func=lambda: _select_from_recent_incidents(pd),
                ),
            ]
        )
    if plugin_manager.config_has_plugin("rootly"):
        rootly = cast("RootlyPlugin", plugin_manager.get_plugin("rootly"))
        options.extend(
            [
                PayloadOption(
                    title="Provide a Rootly incident ID or URL",
                    func=lambda: _provide_rootly_incident_id_or_url(rootly),
                ),
                PayloadOption(
                    title="Select from a list of 20 most recent Rootly incidents",
                    func=lambda: _select_from_recent_rootly_incidents(rootly),
                ),
            ]
        )

    if agent.test_payloads:
        options.insert(
            2,
            PayloadOption(
                title="Run with a test JSON payload from the agent definition",
                func=lambda: _use_test_payload_from_agent(agent),
            ),
        )

    options.append(
        PayloadOption(
            title="Provide the JSON payload directly",
            func=_provide_json_directly,
        )
    )

    selected_option = await select(
        "How would you like to provide the incident?",
        choices=[Choice(opt.title, value=i) for i, opt in enumerate(options)],
    )
    payload = await options[int(selected_option)].func()
    await e.send(
        "payload option selected",
        {"option": options[int(selected_option)].title, "payload_provided": bool(payload)},
    )

    if not payload:
        rich.print("[yellow]No payload provided, skipping the demo.[/yellow]")
        return

    try:
        agent = load_agent(agent.name)
        analysis_agent = AnalysisAgent()
        rich.print("")
        rich.print("Details of the payload we're going to demo:")
        rich.print("-" * 80)
        json_payload = payload
        with contextlib.suppress(json.JSONDecodeError):
            json_payload = json.dumps(json.loads(payload), indent=2)
        payload_lines = json_payload.splitlines()
        if len(payload_lines) > 20:
            display_payload = "...(snipped)...\n" + "\n".join(payload_lines[-20:])
        else:
            display_payload = json_payload
        rich.print(display_payload)

        rich.print("")
        rich.print("> Ready to run the demo agent on this payload?")
        rich.print("")
        await questionary.press_any_key_to_continue().unsafe_ask_async()

        rich.print("> Computing status update... (this may take a minute!)")
        console = Console()
        with console.status("working...", spinner="dots") as status:
            result = await analysis_agent.acall(payload=payload, agent=agent)
            status.update("Done 🎉")
        await e.send("demo complete", {"agent_name_hash": hash_value(agent.name)})

        rich.print("")
        _panel("✅ Agent run complete!")
        rich.print(result)
        rich.print("")
        rich.print("")
        rich.print("You can re-run this demo at any point with:")
        rich.print("")
        rich.print(f"  [bold deep_sky_blue1]unpage agent run {agent.name}[/bold deep_sky_blue1]")
        rich.print("")
    except Exception as ex:
        rich.print(f"[red] Demo failed:[/red] {ex}")
        sys.exit(1)
    await questionary.press_any_key_to_continue(
        message="Ready to wrap up quickstart?"
    ).unsafe_ask_async()
    _panel("🎉 You did it! Next steps")
    rich.print(
        "🎉 You did it! Don't stop now—what do you want to do next? Here are some suggestions below; use the arrow keys to move through each one and see a description."
    )
    rich.print("")
    _ = await select(
        message="Use the arrows to move through the options",
        choices=[
            Choice(
                title="Edit the agent you just ran, or try a different one",
                value="edit_agents",
                description="""You can run `unpage agent -h` to see the full list of available agent commands as you continue to build out your agents



""",
            ),
            Choice(
                title="Learn how to deploy your agent remotely",
                value="learn_to_deploy",
                description="""There are multiple options for deploying your agents. We recommend starting with Guide to Deploying Agents on our docs site. https://docs.unpage.ai/?utm_source=unpage



""",
            ),
            Choice(
                title="Configure more plugins & tools, and build a knowledge graph of your infrastructure",
                value="configure_more",
                description="""Unpage supports a rich infrastructure knowledge graph builder, which can provide helpful context to your Unpage Agents. The graph can be built from your infrastructure tools (like AWS or Aptible) and your observability tools (like Datadog and CloudWatch).

For a full list of our currently-supported Plugins, check out our docs. https://docs.unpage.ai/?utm_source=unpage

If you want to try configuring some more plugins and building the graph, you can run `unpage configure`""",
            ),
        ],
    )
    rich.print(
        "Don't forget to join the Slack community if you haven't already. The Unpage team is always available to answer questions, and you'll be among the first to hear about new updates!"
    )
    rich.print("https://docs.unpage.ai/#learn-more?utm_source=unpage")
    rich.print("")
    rich.print("📖 Docs are at https://docs.unpage.ai/?utm_source=unpage")
    rich.print("")



import os
from datetime import UTC, datetime, timedelta
from typing import Any

import anyio
import questionary
from pydantic import AwareDatetime

from unpage.config import PluginSettings
from unpage.knowledge import Graph
from unpage.plugins.base import Plugin
from unpage.plugins.datadog.client import DatadogClient
from unpage.plugins.datadog.models import DatadogLogSearchResult
from unpage.plugins.datadog.nodes.datadog_api import DatadogApi
from unpage.plugins.datadog.nodes.datadog_service import DatadogService
from unpage.plugins.datadog.nodes.datadog_system import DatadogSystem
from unpage.plugins.datadog.nodes.datadog_team import DatadogTeam
from unpage.plugins.mixins import KnowledgeGraphMixin, tool
from unpage.plugins.mixins.mcp import McpServerMixin

RESULT_LIMIT = 75 * 1024


class DatadogPlugin(Plugin, KnowledgeGraphMixin, McpServerMixin):
    """A plugin for Datadog."""

    _client: DatadogClient

    def __init__(
        self,
        *args: Any,
        api_key: str | None = None,
        application_key: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._api_key = api_key
        self._application_key = application_key
        self._client = DatadogClient(
            api_key=api_key or os.getenv("DATADOG_API_KEY", ""),
            application_key=application_key or os.getenv("DATADOG_APP_KEY", ""),
        )

    async def interactive_configure(self) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        return {
            "api_key": await questionary.password(
                "Enter your Datadog API key",
                default=self._api_key or os.getenv("DATADOG_API_KEY", ""),
                instruction="Generate an API key with https://docs.unpage.ai/plugins/datadog#prerequisites",
            ).unsafe_ask_async(),
            "application_key": await questionary.password(
                "Enter your Datadog application key",
                default=self._application_key or os.getenv("DATADOG_APP_KEY", ""),
                instruction="Generate an application key with https://docs.unpage.ai/plugins/datadog#prerequisites",
            ).unsafe_ask_async(),
        }

    async def populate_graph(self, graph: Graph) -> None:
        """Populate the graph with Datadog entities."""
        async with anyio.create_task_group() as tg:
            tg.start_soon(self._populate_teams, graph)
            tg.start_soon(self._populate_software_catalog_entities, graph)

    async def _populate_teams(self, graph: Graph) -> None:
        async for team in self._client.iter_teams():
            await graph.add_node(
                DatadogTeam(
                    node_id=f"team:default/{team['attributes']['handle']}",
                    raw_data=team.to_dict(),
                    _graph=graph,
                )
            )

    async def _populate_software_catalog_entities(self, graph: Graph) -> None:
        async for entity in self._client.iter_software_catalog_entities():
            entity_kind = entity["attributes"]["kind"]

            if entity_kind == "service":
                node = DatadogService(
                    node_id=f"service:default/{entity['attributes']['name']}",
                    raw_data=entity.to_dict(),
                    _graph=graph,
                )
            elif entity_kind == "api":
                node = DatadogApi(
                    node_id=f"api:default/{entity['attributes']['name']}",
                    raw_data=entity.to_dict(),
                    _graph=graph,
                )
            else:
                raise ValueError(f"Unknown entity kind: {entity_kind}")

            await graph.add_node(node)

            # HACK: There doesn't seem to be a way to list systems directly in datadog, so we'll fake it for now.
            for (
                related_entity_reference,
                _,
            ) in await node._get_related_entity_reference_identifiers():
                if related_entity_reference.startswith("system:"):
                    await graph.add_node(
                        DatadogSystem(
                            node_id=related_entity_reference,
                            raw_data={
                                "attributes": {
                                    "name": related_entity_reference.split("/")[-1],
                                },
                            },
                            _graph=graph,
                        )
                    )

    @tool()
    async def search_logs(
        self,
        query: str,
        min_time: AwareDatetime,
        max_time: AwareDatetime,
    ) -> DatadogLogSearchResult | str:
        """Search Datadog for logs within a given time range

        Args:
            query (str): The search query.
            min_time (AwareDatetime): The starting time for the range within which to search.
            max_time (AwareDatetime): The ending time for the range within which to search.

        Returns:
            DatadogLogSearchResult: logs that matched the query and fit within response limit
        """
        if min_time >= max_time:
            return f"min_time must come before max_time {min_time=} {max_time=}"
        logs = []
        truncated = False
        content_length = 0

        class TimeoutTracker:
            timed_out: bool = False
            start_time: AwareDatetime = datetime.now(UTC)

            def under_time_out(self, _: AwareDatetime | None) -> bool:
                self.timed_out = (datetime.now(UTC) - self.start_time) > timedelta(seconds=10)
                return not self.timed_out

        tt = TimeoutTracker()
        async for log in self._client.search_logs(
            query=query if ":" in query else f"*:*{query}*",
            min_time=min_time,
            max_time=max_time,
            continue_search=tt.under_time_out,
        ):
            content_length += len(log.model_dump_json(indent=6)) + 8
            if content_length >= RESULT_LIMIT:
                truncated = True
                break
            logs.append(log)
        return DatadogLogSearchResult(
            results=logs,
            truncated=truncated or tt.timed_out,
        )



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any

import dspy
from fastmcp import Client, FastMCP
from pydantic import BaseModel, Field
from pydantic_yaml import parse_yaml_file_as

from unpage.config import EnvironmentVariablesMixin, PluginConfig, manager
from unpage.knowledge.graph import Graph
from unpage.mcp import Context, build_mcp_server
from unpage.plugins.base import REGISTRY, PluginManager
from unpage.utils import wildcard_or_regex_match_any


class PartialConfigForAgent(EnvironmentVariablesMixin, BaseModel):
    plugins: dict[str, PluginConfig] = Field(default_factory=dict)


class AgentTestPayload(BaseModel):
    description: str | None = Field(
        description="A description of the test payload and what it is testing",
        default=None,
    )
    payload: str = Field(description="The test payload to use")


class Agent(BaseModel):
    name: str = Field(description="The name of the agent", default="")
    description: str = Field(description="A description of the agent and when it should be used")
    prompt: str = Field(description="The prompt to use for the agent")
    tools: list[str] = Field(description="The tools the agent has access to")
    config: PartialConfigForAgent = Field(
        description="Agent specific configuration to add to, or even override, the global config",
        default_factory=PartialConfigForAgent,
    )
    test_payloads: dict[str, AgentTestPayload] | None = Field(
        description="Test payloads for testing the agent",
        default=None,
    )

    def required_plugins_from_tools(self) -> list[str]:
        allowed_tool_patterns = (
            ["*"]
            if not self.tools
            else list({f"{p.split('_', maxsplit=1)[0]}*" for p in self.tools})
        )
        return [
            plugin_name
            for plugin_name in REGISTRY
            if wildcard_or_regex_match_any(allowed_tool_patterns, f"{plugin_name}_")
        ]


class SelectAgent(dspy.Signature):
    """You are an expert Site Reliability Engineer with deep expertise in alert triage.

    You will be given an alert payload from an alerting system and a list of
    agent descriptions. Your task is to select the most relevant agent to use to
    analyze the given alert. You may use the tools you're given to get more
    information about the alert and the incident, if necessary.

    Note that you MUST select an agent from the list of available agents. If you
    don't know which agent to use, select the "default" agent.
    """

    payload: str = dspy.InputField(
        description="The payload received from the alerting system",
    )

    available_agents: dict[str, Agent] = dspy.InputField(
        description="The list of agents to select from",
    )

    selected_agent_name: str = dspy.OutputField(
        description="The selected agent",
    )

    reasoning: str = dspy.OutputField(
        description="The reasoning behind choosing the selected agent",
    )


class Analyze(dspy.Signature):
    payload: str = dspy.InputField(description="The alert payload to triage")
    analysis: str = dspy.OutputField(description="The triage analysis of the alert payload")


class AnalysisAgent(dspy.Module):
    def __init__(self) -> None:
        super().__init__()

        self.profile = manager.get_active_profile()
        self.config_dir = manager.get_active_profile_directory()
        self.config = manager.get_active_profile_config()
        self.llm_settings = self.config.plugins["llm"].settings

        self.available_agents: dict[str, Agent] = {}
        for agent_file in self.config_dir.glob("agents/**/*.yaml"):
            try:
                agent = parse_yaml_file_as(Agent, agent_file)
            except Exception as ex:
                print(f"Failed to load agent {agent_file}: {ex}")
                continue
            agent.name = agent_file.stem
            self.available_agents[agent.name] = agent

        self.mcp_server = None

    async def get_mcp_server(self, agent: Agent | None) -> FastMCP:
        if self.mcp_server is None:
            config = (
                self.config
                if agent is None or agent.config is None
                else self.config.merge_plugins(agent.config.plugins)
            )
            self.mcp_server = await build_mcp_server(
                Context(
                    profile=self.profile,
                    config=config,
                    plugins=PluginManager(config=config),
                    graph=Graph(self.config_dir / "graph.json"),
                )
            )
        return self.mcp_server

    async def acall(self, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        params = {
            "model": self.llm_settings["model"],
            "api_key": self.llm_settings["api_key"],
            **(
                {"temperature": self.llm_settings["temperature"]}
                if not self.llm_settings["model"].startswith("bedrock/")
                else {}
            ),
            "max_tokens": self.llm_settings["max_tokens"],
            "cache": self.llm_settings["cache"],
        }
        with dspy.context(
            lm=dspy.LM(**params),
        ):
            return await super().acall(*args, **kwargs)

    async def aforward(
        self,
        payload: str,
        agent: Agent | None = None,
        route_only: bool = False,
        max_iters: int = 5,
    ) -> str:
        """Triage the given alert payload using an appropriate prompt."""
        if agent is None:
            selected_agent, reasoning = await self.select_agent(payload)
        else:
            selected_agent, reasoning = agent, "(explicitly selected)"

        if route_only:
            return f"Routing to the {selected_agent.name!r} agent\n  - Agent description: {selected_agent.description}\n  - Selection reasoning: {reasoning}"

        return await self.analyze(
            payload=payload,
            agent=selected_agent,
            max_iters=max_iters,
        )

    async def analyze(
        self,
        payload: str,
        agent: Agent,
        max_iters: int = 5,
    ) -> str:
        """Triage the given alert payload using the selected agent."""
        # Inject the selected prompt into the signature.
        signature = Analyze.with_instructions(agent.prompt)

        async with self.unpage_agent(signature, agent, max_iters=max_iters) as unpage:
            result = await unpage.acall(payload=payload)
            return result.analysis

    async def select_agent(
        self,
        payload: str,
    ) -> tuple[Agent, str]:
        """Select the most relevant prompt to use to triage the given alert payload.

        Returns the selected agent and the reasoning behind the selection.
        """
        async with self.unpage_agent(SelectAgent) as prompt_selector:
            result = await prompt_selector.acall(
                payload=payload,  # type: ignore[reportCallIssue]
                # Pass only prompt names and descriptions to avoid confusing the LLM
                # with the prompt contents.
                available_agents={
                    n: {"name": n, "description": p.description}
                    for n, p in self.available_agents.items()
                },
            )
            return self.available_agents[result.selected_agent_name], result.reasoning

    @asynccontextmanager
    async def unpage_agent(
        self,
        signature: type[dspy.Signature],
        agent: Agent | None = None,
        max_iters: int = 5,
    ) -> AsyncGenerator[dspy.Module, None]:
        """Yield a configured Unpage agent."""
        allowed_tool_patterns = ["*"] if agent is None or not agent.tools else agent.tools

        async with Client(await self.get_mcp_server(agent)) as client:
            yield dspy.ReAct(
                signature,
                tools=[
                    dspy.Tool.from_mcp_tool(client.session, tool)
                    for tool in await client.list_tools()
                    if wildcard_or_regex_match_any(allowed_tool_patterns, tool.name)
                ],
                max_iters=max_iters,
            )



import os
import signal
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import asynccontextmanager

import anyio
import sentry_sdk
from fastmcp import FastMCP
from mcp import ServerResult, types
from pydantic import BaseModel, ConfigDict
from pydantic_core import to_jsonable_python

from unpage.config import Config, manager
from unpage.knowledge import Graph
from unpage.plugins import PluginManager
from unpage.plugins.mixins.mcp import McpServerMixin
from unpage.telemetry import client as telemetry
from unpage.utils import print


class Context(BaseModel, extra="allow"):
    """Context for the MCP server."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    profile: str
    config: Config
    plugins: PluginManager
    mcp_server: FastMCP | None = None
    graph: Graph


async def build_mcp_server(context: Context) -> FastMCP:
    @asynccontextmanager
    async def lifespan(server: FastMCP) -> AsyncIterator[Context]:
        """Initialize the application context and plugins."""
        yield context

    mcp = FastMCP("unpage", lifespan=lifespan)

    def with_telemetry(
        handler: Callable[[types.Request], Awaitable[ServerResult]],
    ) -> Callable[[types.Request], Awaitable[ServerResult]]:
        async def _telemetry_handler(request: types.Request) -> ServerResult:
            await telemetry.send_event(
                {
                    "event": "mcp_request",
                    "request": to_jsonable_python(request),
                }
            )

            try:
                return await handler(request)
            except Exception as e:
                sentry_sdk.capture_exception(e)
                raise e

        return _telemetry_handler

    # Instrument the MCP server with Sentry
    mcp._mcp_server.request_handlers = {
        event_type: with_telemetry(handler)
        for event_type, handler in mcp._mcp_server.request_handlers.items()
    }

    for plugin in context.plugins:
        plugin.context = context

    for mcp_plugin in context.plugins.get_plugins_with_capability(McpServerMixin):
        mcp_plugin.register_all(mcp, prefix=mcp_plugin.name)

    return mcp


async def start(
    disable_stdio: bool = False,
    disable_http: bool = False,
    http_host: str = "127.0.0.1",
    http_port: int = 8000,
) -> None:
    """Start the MCP server.

    This function will start the MCP server with the specified transports.

    Args:
        disable_sse: If True, the SSE transport will be disabled.
        disable_stdio: If True, the stdio transport will be disabled.
    """
    # Register direct signal handlers for immediate exit
    signal.signal(signal.SIGINT, lambda *args: os._exit(0))
    signal.signal(signal.SIGTERM, lambda *args: os._exit(0))

    config = manager.get_active_profile_config()
    plugins = PluginManager(config=config)
    context = Context(
        profile=manager.get_active_profile(),
        config=config,
        plugins=plugins,
        graph=Graph(manager.get_active_profile_directory() / "graph.json"),
    )

    mcp = await build_mcp_server(context)

    async def _run_stdio_server() -> None:
        await mcp.run_stdio_async(show_banner=False)

    async def _run_http_server() -> None:
        await mcp.run_http_async(
            show_banner=False,
            transport="http",
            host=http_host,
            port=http_port,
        )

    async with anyio.create_task_group() as tg:
        if not disable_stdio:
            tg.start_soon(_run_stdio_server)
        if not disable_http:
            tg.start_soon(_run_http_server)

    print("MCP server stopped")



from aiobotocore.config import AioConfig
from botocore.exceptions import ClientError

from unpage.plugins.aws.arn.arn import AwsArn
from unpage.plugins.aws.nodes.base import AwsNode
from unpage.plugins.aws.utils import swallow_boto_client_access_errors
from unpage.utils import print


class AwsAlbTargetGroup(AwsNode):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["TargetGroupArn"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *((lb, "routes_requests_for") for lb in self.raw_data.get("LoadBalancerArns", [])),
            self.raw_data.get("VpcId"),
            *await self._get_targets(),
        ]

    async def _get_targets(self) -> list[tuple[str, str]]:
        try:
            arn = AwsArn.parse(self.node_id)
        except ValueError as ex:
            print(f"Invalid ARN format for target group {self.node_id}: {ex}")
            return []
        async with (
            swallow_boto_client_access_errors(service_name="elbv2", region=arn.region),
            self.session.client(
                "elbv2",
                region_name=arn.region,
                config=AioConfig(retries={"max_attempts": 10, "mode": "adaptive"}),
            ) as client,
        ):
            try:
                resp = await client.describe_target_health(TargetGroupArn=self.node_id)
            except ClientError as ex:
                print(f"Error describing target health for {self.node_id}: {ex}")
                return []
        # FIXME: find a way to scope this to only searching aws resources (or maybe even ec2 resources) [tvd, 2025-06-27]
        return [
            (target.get("Target", {}).get("Id"), "has_target")
            for target in resp.get("TargetHealthDescriptions", [])
            if target.get("Target", {}).get("Id") is not None
        ]



import shutil
from typing import Any

from fastmcp import Client, FastMCP
from fastmcp.mcp_config import MCPConfig, MCPServerTypes, StdioMCPServer

from unpage.plugins.base import Plugin
from unpage.plugins.mcp.transport import CompositeMCPTransport
from unpage.plugins.mixins.mcp import McpServerMixin
from unpage.utils import suppress_output


class McpProxyPlugin(Plugin, McpServerMixin):
    """A plugin that proxies MCP requests to a remote MCP server."""

    abstract = True
    prefix_tools: bool = True
    default_enabled: bool = False

    async def validate_plugin_config(self) -> None:
        """Validate the plugin config."""
        config = self.get_mcp_config()
        for name, server in config.mcpServers.items():
            # Validate that the configured command is in the PATH
            if isinstance(server, StdioMCPServer) and not shutil.which(server.command):
                raise ValueError(f"Command {server.command!r} not found in PATH")

            # Attempt to connect to the server as a simple auth check.
            #
            # Note, however, than many MCP servers will allow you to connect
            # even with invalid credentials, and may only throw auth-related
            # errors when calling tools or accessing resources.
            #
            # Individual plugins are encouraged to override this method and
            # provide additional validation.
            with suppress_output():
                try:
                    async with Client(
                        self.get_mcp_server(MCPConfig(mcpServers={name: server}))
                    ) as client:
                        if not client.is_connected():
                            raise ValueError(
                                f"Unable to connect to server {name!r}. Are your credentials correct?"
                            )
                except Exception as ex:
                    raise ValueError(f"Error connecting to server {name!r}: {ex}") from ex

    def get_mcp_server_settings(self) -> MCPServerTypes:
        """Return an MCP server settings for the proxy server."""
        raise NotImplementedError("get_mcp_server_settings must be implemented by subclasses")

    def get_mcp_config(self) -> MCPConfig:
        """Return an MCP config for the proxy server."""
        return MCPConfig(mcpServers={self.name: self.get_mcp_server_settings()})

    def get_mcp_server(self, config: MCPConfig | None = None) -> FastMCP[Any]:
        """Return an MCP proxy server for the given config."""
        config = config or self.get_mcp_config()

        if not config.mcpServers:
            return super().get_mcp_server()

        return FastMCP.as_proxy(
            backend=Client(
                transport=CompositeMCPTransport(
                    config=config,
                    name_as_prefix=self.prefix_tools,
                ),
            ),
        )

    def get_mcp_client(self, config: MCPConfig | None = None) -> Client[Any]:
        """Return an MCP client for the given config."""
        return Client(self.get_mcp_server(config or self.get_mcp_config()))

    async def list_tools(self) -> list[str]:
        """List the tools available on the MCP server."""
        async with self.get_mcp_client() as client:
            return [tool.name for tool in await client.list_tools()]

    async def call_tool(self, tool_name: str, *args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
        """Call an MCP tool."""
        async with self.get_mcp_client() as client:
            return await client.call_tool(tool_name, *args, **kwargs)


class McpPlugin(McpProxyPlugin):
    mcp_servers: dict[str, MCPServerTypes]

    def __init__(
        self,
        *args: Any,
        mcp_servers: dict[str, MCPServerTypes] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.mcp_servers = mcp_servers or {}

    def get_mcp_config(self) -> MCPConfig:
        return MCPConfig(mcpServers=self.mcp_servers)



import os
from collections.abc import AsyncGenerator
from typing import Any

import questionary
import rich
from pydantic import AwareDatetime, BaseModel

from unpage.config import PluginSettings
from unpage.knowledge import Graph
from unpage.plugins.base import Plugin
from unpage.plugins.mixins import KnowledgeGraphMixin, McpServerMixin, tool
from unpage.plugins.rootly.client import RootlyClient
from unpage.plugins.rootly.models import RootlyIncident, RootlyIncidentPayload
from unpage.utils import classproperty


class RootlyPluginSettings(BaseModel):
    api_key: str = ""


class RootlyPlugin(Plugin, KnowledgeGraphMixin, McpServerMixin):
    """A plugin for Rootly."""

    _client: RootlyClient

    def __init__(
        self,
        *args: Any,
        api_key: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._api_key = api_key or os.getenv("ROOTLY_API_KEY", "")
        self._client = RootlyClient(api_key=self._api_key)

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return RootlyPluginSettings(
            api_key=os.environ.get("ROOTLY_API_KEY", ""),
        ).model_dump()

    async def interactive_configure(self) -> PluginSettings:
        """Interactive wizard for configuring the settings of this plugin."""
        defaults = self.default_plugin_settings
        rich.print(
            "[bold]Rootly Plugin Configuration[/bold]: This plugin enables interaction with Rootly incidents. You'll need an API key from your Rootly organization settings."
        )
        rich.print("")
        return {
            "api_key": await questionary.password(
                "API Key",
                default=self._api_key or defaults.get("api_key", ""),
                instruction="Generate a token from Organization Settings > API Keys in Rootly",
            ).unsafe_ask_async(),
        }

    async def validate_plugin_config(self) -> None:
        if not self._api_key:
            raise LookupError("plugins.rootly.settings.api_key (or $ROOTLY_API_KEY) must be set")
        # Test API connectivity
        await self._client.list_incidents({"page[size]": "1"})

    @tool()
    async def list_recent_incidents(self, limit: int = 10) -> list[dict[str, Any]]:
        """List recent Rootly incidents

        Args:
            limit (int): Maximum number of incidents to return (default: 10)

        Returns:
            list[dict]: List of recent incidents with their details
        """
        params = {
            "sort": "-created_at",
            "page[size]": str(min(limit, 100)),  # API limit
        }
        response = await self._client.list_incidents(params)
        return response.get("data", [])

    @tool()
    async def get_incident_details(self, incident_id: str) -> dict[str, Any]:
        """Get a Rootly incident by ID, including all incident details.

        Args:
            incident_id (str): The ID of the Rootly incident. Typically a
            UUID string. For example "01234567-89ab-cdef-0123-456789abcdef".

        Returns:
            dict: The incident JSON payload, including all incident details.
        """
        return await self._client.get_incident(incident_id)

    @tool()
    async def get_alert_details_for_incident(self, incident_id: str) -> list[dict[str, Any]]:
        """Get the details of the alert(s) for a Rootly incident.

        Args:
            incident_id (str): The ID of the Rootly incident. Typically a
            UUID string. For example "01234567-89ab-cdef-0123-456789abcdef".

        Returns:
            list[dict]: The list of alert details (incident events).
        """
        response = await self._client.get_incident_events(incident_id)
        return response.get("data", [])

    @tool()
    async def post_status_update(
        self, incident_id: str, message: str, visibility: str = "external"
    ) -> None:
        """Post a status update to a Rootly incident

        Args:
            incident_id (str): The Rootly ID of the incident
            message (str): The message to post
            visibility (str): Event visibility - "internal" or "external" (default: "external")
        """
        await self.add_incident_event(incident_id, f"Status Update: {message}", visibility)

    @tool()
    async def add_incident_event(
        self, incident_id: str, event_description: str, visibility: str = "internal"
    ) -> dict[str, Any]:
        """Add a timeline event to a Rootly incident for root cause analysis

        Args:
            incident_id (str): The Rootly ID of the incident
            event_description (str): Description of the event/activity
            visibility (str): Event visibility - "internal" or "external" (default: "internal")

        Returns:
            dict: The created event data

        Note: Creates structured timeline entries for proper incident tracking
        and root cause analysis. Events are timestamped automatically by Rootly.
        """
        event_data = {
            "data": {
                "type": "incident_events",
                "attributes": {
                    "event": event_description,
                    "kind": "event",
                    "source": "api",
                    "visibility": visibility,
                },
            }
        }

        return await self._client.create_incident_event_new(incident_id, event_data)

    @tool()
    async def log_investigation_finding(
        self,
        incident_id: str,
        finding: str,
        source: str | None = None,
        visibility: str = "internal",
    ) -> dict[str, Any]:
        """Log an investigation finding for root cause analysis

        Args:
            incident_id (str): The Rootly ID of the incident
            finding (str): The investigation finding or discovery
            source (str, optional): Source of the finding (e.g., 'logs', 'metrics', 'database')
            visibility (str): Event visibility - "internal" or "external" (default: "internal")

        Returns:
            dict: The created event data
        """
        event_description = f"Investigation Finding: {finding}"
        if source:
            event_description += f" (Source: {source})"

        return await self.add_incident_event(incident_id, event_description, visibility)

    @tool()
    async def log_action_taken(
        self,
        incident_id: str,
        action: str,
        outcome: str | None = None,
        visibility: str = "internal",
    ) -> dict[str, Any]:
        """Log an action taken during incident response

        Args:
            incident_id (str): The Rootly ID of the incident
            action (str): The action that was taken
            outcome (str, optional): The result or outcome of the action
            visibility (str): Event visibility - "internal" or "external" (default: "internal")

        Returns:
            dict: The created event data
        """
        event_description = f"Action Taken: {action}"
        if outcome:
            event_description += f" → Result: {outcome}"

        return await self.add_incident_event(incident_id, event_description, visibility)

    @tool()
    async def log_escalation(
        self,
        incident_id: str,
        escalated_to: str,
        reason: str | None = None,
        visibility: str = "internal",
    ) -> dict[str, Any]:
        """Log an incident escalation

        Args:
            incident_id (str): The Rootly ID of the incident
            escalated_to (str): Who or what team the incident was escalated to
            reason (str, optional): Reason for the escalation
            visibility (str): Event visibility - "internal" or "external" (default: "internal")

        Returns:
            dict: The created event data
        """
        event_description = f"Escalated to {escalated_to}"
        if reason:
            event_description += f" - Reason: {reason}"

        return await self.add_incident_event(incident_id, event_description, visibility)

    @tool()
    async def log_communication(
        self, incident_id: str, communication_type: str, details: str, visibility: str = "external"
    ) -> dict[str, Any]:
        """Log a communication event (e.g., customer notification, internal update)

        Args:
            incident_id (str): The Rootly ID of the incident
            communication_type (str): Type of communication (e.g., "Customer Notification", "Team Update")
            details (str): Details of the communication
            visibility (str): Event visibility - "external" for customer-facing, "internal" for team-only

        Returns:
            dict: The created event data
        """
        event_description = f"{communication_type}: {details}"

        return await self.add_incident_event(incident_id, event_description, visibility)

    @tool()
    async def get_incident_timeline(self, incident_id: str) -> list[dict[str, Any]]:
        """Get the complete timeline of events for an incident

        Args:
            incident_id (str): The Rootly ID of the incident

        Returns:
            list: List of all events/timeline entries for the incident
        """
        response = await self._client.get_incident_events(incident_id)
        return response.get("data", [])

    @tool()
    async def update_incident_event(
        self, event_id: str, event_description: str, visibility: str = "internal"
    ) -> dict[str, Any]:
        """Update an existing incident event

        Args:
            event_id (str): The ID of the event to update
            event_description (str): Updated description of the event
            visibility (str): Event visibility - "internal" or "external" (default: "internal")

        Returns:
            dict: The updated event data
        """
        update_data = {
            "data": {
                "type": "incident_events",
                "id": event_id,
                "attributes": {"event": event_description, "visibility": visibility},
            }
        }

        return await self._client.update_incident_event(event_id, update_data)

    @tool()
    async def delete_incident_event(self, event_id: str) -> dict[str, Any]:
        """Delete an incident event

        Args:
            event_id (str): The ID of the event to delete

        Returns:
            dict: Response from the deletion request
        """
        return await self._client.delete_incident_event(event_id)

    @tool()
    async def resolve_incident(
        self, incident_id: str, resolution_message: str | None = None
    ) -> None:
        """Resolve a Rootly incident

        Args:
            incident_id (str): The ID of the Rootly incident to resolve
            resolution_message (str, optional): A message to include with the resolution
        """
        update_data = {
            "data": {"type": "incidents", "id": incident_id, "attributes": {"status": "resolved"}}
        }

        # Add resolution message if provided
        if resolution_message:
            update_data["data"]["attributes"]["resolution_message"] = resolution_message

        await self._client.update_incident(incident_id, update_data)

    @tool()
    async def mitigate_incident(self, incident_id: str) -> None:
        """Mitigate a Rootly incident

        Args:
            incident_id (str): The ID of the Rootly incident to mitigate
        """
        update_data = {
            "data": {"type": "incidents", "id": incident_id, "attributes": {"status": "mitigated"}}
        }
        await self._client.update_incident(incident_id, update_data)

    @tool()
    async def acknowledge_incident(self, incident_id: str) -> None:
        """Acknowledge a Rootly incident (moves to in_triage status)

        Args:
            incident_id (str): The ID of the Rootly incident to acknowledge

        Note: Sets status to 'in_triage' as this represents acknowledgment
        in the Rootly workflow.
        """
        update_data = {
            "data": {"type": "incidents", "id": incident_id, "attributes": {"status": "in_triage"}}
        }
        await self._client.update_incident(incident_id, update_data)

    async def get_incident_by_id(self, incident_id: str) -> RootlyIncident:
        """Get a single incident by its id

        Args:
            incident_id (str): The Rootly ID of the incident

        Returns:
            RootlyIncident: the rootly incident data
        """
        response = await self._client.get_incident(incident_id)
        return RootlyIncident(**response["data"])

    async def recent_incident_payloads(
        self, since: AwareDatetime | None = None, sort: str = "-created_at"
    ) -> AsyncGenerator[RootlyIncidentPayload, None]:
        """Get a list of recent Rootly incidents.

        Returns:
            AsyncGenerator: incident JSON payloads that can be used for demoing and testing agents
        """
        params = {
            "sort": sort,
            "page[size]": 100,
        }
        if since:
            params["filter[created_at][gte]"] = since.isoformat()

        response = await self._client.list_incidents(params)

        for incident_data in response.get("data", []):
            yield RootlyIncidentPayload(incident=RootlyIncident(**incident_data))

    async def populate_graph(self, graph: Graph) -> None:
        """Populate the graph with Rootly incidents."""
        # Get recent incidents and add them to the graph
        async for incident_payload in self.recent_incident_payloads():
            incident = incident_payload.incident
            from unpage.plugins.rootly.nodes import RootlyIncident

            # Create a Rootly incident node
            incident_node = RootlyIncident(
                node_id=f"rootly_incident:{incident.id}",
                raw_data=incident.model_dump(),
                _graph=graph,
            )
            await graph.add_node(incident_node)



from typing import Annotated

from cyclopts import Parameter

from unpage.agent.app import listen, settings
from unpage.cli.agent._app import agent_app
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@agent_app.command
async def serve(
    *,
    host: str = settings.UNPAGE_HOST,
    port: int = settings.UNPAGE_PORT,
    workers: int = settings.UNPAGE_WORKERS,
    reload: bool = settings.UNPAGE_RELOAD,
    tunnel: bool = settings.UNPAGE_TUNNEL,
    ngrok_token: Annotated[
        str, Parameter(show_default=lambda v: v if not v else f"{v[:3]}<redacted>")
    ] = settings.NGROK_TOKEN,
    ngrok_domain: str = settings.NGROK_DOMAIN,
    debug: bool = settings.UNPAGE_DEBUG,
) -> None:
    """Run the Unpage Agent server, which loads all agents and routes between them. This is intended to be a webhook receiver for PagerDuty.

    Parameters
    ----------
    host
        The host to bind to
    port
        The port to bind to
    workers
        The number of workers to use
    reload
        Reload the server when the code changes
    tunnel
        Tunnel the server through ngrok
    ngrok_token
        The ngrok token to use to tunnel the server
    ngrok_domain
        The ngrok domain to use to tunnel the server
    """
    await telemetry.send_event(
        {
            "command": "agent serve",
            "host": host
            if host.startswith("127")
            else "0.0.0.0"  # noqa: S104 Possible binding to all interfaces
            if host == "0.0.0.0"  # noqa: S104 Possible binding to all interfaces
            else f"{host.split('.')[0]}.0.0.0",
            "port": port,
            "workers": workers,
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "reload": reload,
            "tunnel": tunnel,
        }
    )
    await listen(
        host=host,
        port=port,
        workers=workers,
        tunnel=tunnel,
        ngrok_token=ngrok_token,
        ngrok_domain=ngrok_domain,
        reload=reload,
        debug=debug,
    )



import asyncio
import shlex

from rich import print

from unpage.cli.mlflow._app import mlflow_app
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry
from unpage.utils import confirm


@mlflow_app.command
async def serve(
    *,
    port: int = 5566,
) -> None:
    """Start MLflow tracking server

    Parameters
    ----------
    port
        Port for MLflow server to listen on
    """
    await telemetry.send_event(
        {
            "command": "mlflow serve",
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "port": port,
        }
    )
    host = "127.0.0.1"
    config_dir = manager.get_active_profile_directory()
    mlflow_db = config_dir / "mlflow" / "debug.db"
    mlflow_db.parent.mkdir(parents=True, exist_ok=True)
    mlflow_db.unlink(missing_ok=True)
    backend_store_uri = f"sqlite:///{mlflow_db.absolute()}"
    artifacts_dir = mlflow_db.parent / "mlartifacts"
    artifacts_dir.mkdir(parents=True, exist_ok=True)
    cmd = (
        "uvx mlflow@3 server "
        f"--host {shlex.quote(host)} --port {shlex.quote(str(port))} "
        f"--backend-store-uri {shlex.quote(backend_store_uri)} "
        f"--artifacts-destination {shlex.quote(str(artifacts_dir.absolute()))}"
    )
    print(
        f"MLflow tracking server is ready to start! Set this in your environment to use it: MLFLOW_TRACKING_URI=http://{host}:{port}"
    )
    while not await confirm("Ready to start MLflow tracking server?"):
        pass
    print(f"  ...running: {cmd}")
    proc = await asyncio.create_subprocess_shell(cmd)
    await proc.wait()



from unpage.knowledge import Graph
from unpage.plugins import PluginCapability


class KnowledgeGraphMixin(PluginCapability):
    """Capability for plugins that can add nodes and edges to the graph."""

    async def populate_graph(self, graph: Graph) -> None:
        """Initialize the graph with nodes and edges from the plugin."""
        raise NotImplementedError



from rich import print
from rich.console import Console
from rich.table import Table

from unpage.agent.utils import get_agent_template_description, get_agent_templates
from unpage.cli.agent._app import agent_app
from unpage.telemetry import client as telemetry


@agent_app.command
async def templates() -> None:
    """List the available agent templates."""
    await telemetry.send_event(
        {
            "command": "agent templates",
        }
    )
    print("Available agent templates:")
    table = Table("Template Name", "Description")
    for template in sorted(get_agent_templates()):
        table.add_row(template, get_agent_template_description(template).splitlines()[0])
    console = Console()
    console.print(table)



import colorsys
import importlib
import json
import os
import re
import shlex
import shutil
import subprocess
import sys
from collections.abc import Awaitable, Callable, Generator, Iterable, Sequence
from contextlib import contextmanager, nullcontext, redirect_stderr, redirect_stdout
from io import TextIOWrapper
from pathlib import Path
from re import Pattern
from types import ModuleType
from typing import Any, cast, overload

import anyio
import questionary
import questionary.constants
import rich
from anyio import create_memory_object_stream
from anyio.abc import TaskGroup
from boltons.iterutils import remap
from pydantic import AnyUrl, ValidationError
from rich.console import Console

stderr = Console(stderr=True)


def print(message: str, _indent: int | None = None, **kwargs: Any) -> None:
    stderr.print_json(json.dumps({"message": message, **kwargs}, default=str), indent=_indent)


Choice = questionary.Choice


async def confirm(
    message: str,
    default: bool = True,
    qmark: str = ">",
    style: questionary.Style | None = None,
    auto_enter: bool = False,
    instruction: str | None = None,
    **kwargs: Any,
) -> bool:
    """Convenience wrapper around await questionary.confirm().unsafe_ask_async()"""
    return await questionary.confirm(
        message=message,
        default=default,
        qmark=qmark,
        style=style,
        auto_enter=auto_enter,
        instruction=instruction,
        **kwargs,
    ).unsafe_ask_async()


@overload
async def select(
    message: str,
    choices: Sequence[str],
    default: str | None = None,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    use_shortcuts: bool = False,
    use_arrow_keys: bool = True,
    use_indicator: bool = False,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: bool = False,
    show_selected: bool = False,
    show_description: bool = True,
    instruction: str | None = None,
    **kwargs: Any,
) -> str: ...


@overload
async def select(
    message: str,
    choices: Sequence[questionary.Choice],
    default: questionary.Choice | None = None,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    use_shortcuts: bool = False,
    use_arrow_keys: bool = True,
    use_indicator: bool = False,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: bool = False,
    show_selected: bool = False,
    show_description: bool = True,
    instruction: str | None = None,
    **kwargs: Any,
) -> str: ...


@overload
async def select(
    message: str,
    choices: Sequence[dict[str, Any]],
    default: dict[str, Any] | None = None,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    use_shortcuts: bool = False,
    use_arrow_keys: bool = True,
    use_indicator: bool = False,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: bool = False,
    show_selected: bool = False,
    show_description: bool = True,
    instruction: str | None = None,
    **kwargs: Any,
) -> dict[str, Any]: ...


async def select(
    message: str,
    choices: Sequence[str | questionary.Choice | dict[str, Any]],
    default: str | questionary.Choice | dict[str, Any] | None = None,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    use_shortcuts: bool = False,
    use_arrow_keys: bool = True,
    use_indicator: bool = False,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: bool = False,
    show_selected: bool = False,
    show_description: bool = True,
    instruction: str | None = None,
    **kwargs: Any,
) -> str | questionary.Choice | dict[str, Any]:
    """Convenience wrapper around await questionary.select().unsafe_ask_async()"""
    return await questionary.select(
        message=message,
        choices=choices,
        default=default,
        qmark=qmark,
        pointer=pointer,
        style=style,
        use_shortcuts=use_shortcuts,
        use_arrow_keys=use_arrow_keys,
        use_indicator=use_indicator,
        use_jk_keys=use_jk_keys,
        use_emacs_keys=use_emacs_keys,
        use_search_filter=use_search_filter,
        show_selected=show_selected,
        show_description=show_description,
        instruction=instruction,
        **kwargs,
    ).unsafe_ask_async()


@overload
async def checkbox(
    message: str,
    choices: Sequence[str],
    default: str | None = None,
    validate: Callable[[list[str]], bool | str] = lambda a: True,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    initial_choice: str | None = None,
    use_arrow_keys: bool = True,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: str | bool | None = False,
    instruction: str | None = None,
    show_description: bool = True,
    **kwargs: Any,
) -> Sequence[str]: ...


@overload
async def checkbox(
    message: str,
    choices: Sequence[questionary.Choice],
    default: str | None = None,
    validate: Callable[[list[str]], bool | str] = lambda a: True,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    initial_choice: questionary.Choice | None = None,
    use_arrow_keys: bool = True,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: str | bool | None = False,
    instruction: str | None = None,
    show_description: bool = True,
    **kwargs: Any,
) -> Sequence[questionary.Choice]: ...


@overload
async def checkbox(
    message: str,
    choices: Sequence[dict[str, Any]],
    default: str | None = None,
    validate: Callable[[list[str]], bool | str] = lambda a: True,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    initial_choice: dict[str, Any] | None = None,
    use_arrow_keys: bool = True,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: str | bool | None = False,
    instruction: str | None = None,
    show_description: bool = True,
    **kwargs: Any,
) -> Sequence[dict[str, Any]]: ...


async def checkbox(
    message: str,
    choices: Sequence[str | questionary.Choice | dict[str, Any]],
    default: str | None = None,
    validate: Callable[[list[str]], bool | str] = lambda a: True,
    qmark: str = ">",
    pointer: str | None = questionary.constants.DEFAULT_SELECTED_POINTER,
    style: questionary.Style | None = None,
    initial_choice: str | questionary.Choice | dict[str, Any] | None = None,
    use_arrow_keys: bool = True,
    use_jk_keys: bool = True,
    use_emacs_keys: bool = True,
    use_search_filter: str | bool | None = False,
    instruction: str | None = None,
    show_description: bool = True,
    **kwargs: Any,
) -> Sequence[str | questionary.Choice | dict[str, Any]]:
    return await questionary.checkbox(
        message=message,
        choices=choices,
        default=default,
        validate=validate,
        qmark=qmark,
        pointer=pointer,
        style=style,
        initial_choice=initial_choice,
        use_arrow_keys=use_arrow_keys,
        use_jk_keys=use_jk_keys,
        use_emacs_keys=use_emacs_keys,
        use_search_filter=use_search_filter,
        instruction=instruction,
        show_description=show_description,
        **kwargs,
    ).unsafe_ask_async()


def import_submodules(
    package: str | ModuleType,
    recursive: bool = True,
) -> None:
    """Import all submodules of a module, recursively, including subpackages

    Uses filesystem scanning to avoid namespace pollution from installed packages
    with similar names.

    :param package: package (name or actual module)
    :type package: str | module
    :rtype: dict[str, types.ModuleType]
    """
    if isinstance(package, str):
        package = importlib.import_module(package)

    # Get the actual filesystem modules only to avoid namespace pollution
    package_dir = Path(package.__path__[0])  # Get the first path

    for item_path in package_dir.iterdir():
        # Skip hidden files and private modules.
        if item_path.name[0] in ("_", "."):
            continue

        name = item_path.stem
        is_pkg = item_path.is_dir() and (item_path / "__init__.py").exists()

        # Skip non-Python files and folders.
        if not is_pkg and item_path.suffix != ".py":
            continue

        full_name = f"{package.__name__}.{name}"
        importlib.import_module(full_name)

        if is_pkg and recursive:
            import_submodules(full_name)


class classproperty[T]:  # noqa: N801
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """

    def __init__(self, method: Callable[[Any], T]) -> None:
        self.fget = method

    def __get__(self, instance: object, cls: type | None = None) -> T:
        return self.fget(cls if cls else instance.__class__)

    def getter(self, method: Callable[[Any], T]) -> "classproperty[T]":
        self.fget = method
        return self


def as_completed[T](tg: TaskGroup, aws: Iterable[Awaitable[T]]) -> Iterable[Awaitable[T]]:
    send_stream, receive_stream = create_memory_object_stream[T | Exception]()

    # Convert the iterable to a list to get its length
    aws_list = list(aws)
    count = len(aws_list)
    completed = 0

    async def populate_result(a: Awaitable[T]) -> None:
        nonlocal completed
        try:
            result = await a
            await send_stream.send(result)
        except Exception as e:
            # Send the exception too, so it can be raised in the caller's context
            await send_stream.send(e)
        finally:
            completed += 1
            # Close the send stream when all tasks are done
            if completed >= count:
                await send_stream.aclose()

    async def wait_for_result() -> T:
        try:
            result = await receive_stream.receive()
            # If we received an exception, raise it
            if isinstance(result, Exception):
                raise result
            return result
        except anyio.EndOfStream as e:
            # This should only happen if all senders are done but we're still trying to receive
            raise StopIteration("No more results available") from e

    for a in aws_list:
        tg.start_soon(populate_result, a)

    return (wait_for_result() for _ in aws_list)


def generate_contrasting_colors(n: int) -> list[tuple[str, str]]:
    """
    Generate N visually distinct colors as (background, foreground) hex pairs.

    The background colors are chosen by evenly spacing hues around the HSL color wheel.
    The foreground color is either black or white, chosen for maximum contrast.

    Args:
        n (int): Number of color pairs to generate.

    Returns:
        List[Tuple[str, str]]: List of (background_hex, foreground_hex) tuples.
    """

    def luminance(r: float, g: float, b: float) -> float:
        # sRGB to linear RGB conversion
        def to_linear(c: float) -> float:
            return c / 12.92 if c <= 0.03928 else ((c + 0.055) / 1.055) ** 2.4

        r_lin, g_lin, b_lin = map(to_linear, (r, g, b))
        return 0.2126 * r_lin + 0.7152 * g_lin + 0.0722 * b_lin

    def get_foreground_color(r: float, g: float, b: float) -> str:
        return "#000000" if luminance(r, g, b) > 0.5 else "#ffffff"

    color_pairs: list[tuple[str, str]] = []
    for i in range(n):
        hue = i / n
        saturation = 0.65
        lightness = 0.5
        r, g, b = colorsys.hls_to_rgb(hue, lightness, saturation)
        hex_color = f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"
        fg_color = get_foreground_color(r, g, b)
        color_pairs.append((hex_color, fg_color))

    return color_pairs


def camel_to_snake(s: str) -> str:
    """Convert a camelCase string to a snake_case string."""
    return re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()


# Define a constant for redacted content
REDACTED = "REDACTED"


def strip_secrets(data: dict[str, Any]) -> dict[str, Any]:
    """Strip secrets from a dictionary."""

    def _visit(_path: tuple[str, ...], key: str, value: Any) -> bool | tuple[str, Any]:  # noqa: ANN401
        if key in ("passphrase", "connection_url"):
            return False
        if key == "env" and isinstance(value, dict):
            value = cast("dict[str, Any]", value)
            clean_env: dict[str, Any] = {}
            for k, v in value.items():
                clean_env[k] = v

                # If it's a URL, redact the password
                try:
                    original_url = AnyUrl(v)
                except ValidationError:
                    pass
                else:
                    if not original_url.host:
                        clean_env[k] = str(original_url)
                        continue
                    else:
                        clean_env[k] = str(
                            AnyUrl.build(
                                scheme=original_url.scheme,
                                username=original_url.username,
                                password=REDACTED,
                                host=original_url.host,
                                port=original_url.port,
                                path=original_url.path,
                                query=original_url.query,
                                fragment=original_url.fragment,
                            )
                        )

                if any(
                    secret_indicator in k.upper()
                    for secret_indicator in (
                        "KEY",
                        "PASS",
                        "PASSPHRASE",
                        "PASSWORD",
                        "PSK",
                        "SECRET",
                        "TOKEN",
                    )
                ):
                    clean_env[k] = REDACTED
            return "env", clean_env
        return True

    return remap(data, _visit)


def compile_regex(pattern: str) -> Pattern[str]:
    """Compile a string into a regex pattern."""
    # Extract the pattern and flags from the string
    delimiter = pattern[0]
    if pattern.rfind(delimiter) == 0:
        raise ValueError("Invalid regex pattern: Missing closing delimiter")

    # Split the pattern and the flags
    parts = pattern.rsplit(delimiter, 1)
    raw_pattern = parts[0][1:]
    flags_str = parts[1] if len(parts) > 1 else ""

    # Map string flags to re module flags
    flag_map = {
        "g": 0,  # Global flag is not needed in Python regex
        "i": re.IGNORECASE,
        "m": re.MULTILINE,
        "s": re.DOTALL,
        "x": re.VERBOSE,
        "u": re.UNICODE,
    }

    # Calculate the combined flags
    flags = 0
    for char in flags_str:
        if char in flag_map:
            flags |= flag_map[char]
        else:
            raise ValueError(f"Unknown regex flag: {char}")

    # Compile the regex pattern with the calculated flags
    return re.compile(raw_pattern, flags)


def wildcard_or_regex_match(pattern: str, string: str) -> bool:
    """Match a string against a wildcard or regex pattern.

    - If pattern is surrounded by slashes (/regex/), treat it as a raw regular expression.
    - Otherwise, treat '*' as a wildcard matching any sequence of characters.
    """
    if pattern.startswith("/") and pattern.endswith("/") and len(pattern) > 1:
        # Regex mode
        regex = pattern[1:-1]  # Strip the surrounding slashes
    else:
        # Wildcard mode: escape and convert '*' to '.*'
        escaped = re.escape(pattern)
        regex = "^" + escaped.replace(r"\*", ".*") + "$"

    return re.match(regex, string) is not None


def wildcard_or_regex_match_any(patterns: list[str], string: str) -> bool:
    """Match a string against any of the given wildcard or regex patterns."""
    return any(wildcard_or_regex_match(pattern, string) for pattern in patterns)


async def edit_file(file_path: str | Path, editor: str | None = None) -> None:
    """Edit a file in the user's editor."""
    editor = editor or get_editor()
    if not editor:
        raise ValueError("No editor specified")

    # Split editor command into command and arguments
    editor_cmd, *editor_args = shlex.split(editor)

    # Check if the editor command exists in PATH
    editor_path = shutil.which(editor_cmd)
    if not editor_path:
        rich.print(f"[red]Editor '{editor_cmd}' not found in $PATH[/red]")
        rich.print(f"[blue]Please manually open: {str(file_path)!r}[/blue]")
        return

    try:
        # Combine editor path, args, and file path into the command
        command = [editor_path, *editor_args, str(file_path)]
        await anyio.run_process(
            command,
            # Set the standard streams so that the editor process can properly interact with the terminal.
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
    except subprocess.CalledProcessError:
        rich.print(f"[red]Failed to open {str(file_path)!r} with {editor!r}[/red]")
        rich.print(f"[blue]Please manually open: {str(file_path)!r}[/blue]")
    except FileNotFoundError:
        rich.print(f"[red]Editor '{editor_cmd}' not found[/red]")
        rich.print(f"[blue]Please manually open: {str(file_path)!r}[/blue]")


def get_editor() -> str | None:
    return os.environ.get("DAYDREAM_EDITOR") or os.environ.get("EDITOR") or _get_default_editor()


def _get_default_editor() -> str | None:
    if vim := shutil.which("vim"):
        return vim
    if vi := shutil.which("vi"):
        return vi
    if nano := shutil.which("nano"):
        return nano
    if pico := shutil.which("pico"):
        return pico
    if emacs := shutil.which("emacs"):
        return emacs


@contextmanager
def suppress_output(
    stdout: bool = True, stderr: bool = True
) -> Generator[tuple[TextIOWrapper | None, TextIOWrapper | None], None, None]:
    with (
        Path(os.devnull).open("w") as devnull,
        redirect_stdout(devnull) if stdout else nullcontext() as out,
        redirect_stderr(devnull) if stderr else nullcontext() as err,
    ):
        yield (out, err)



import sys

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import (
    cleanup_pid_file,
    get_log_file,
    get_pid_file,
    is_process_running,
)
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command
async def status() -> None:
    """Check if graph build is running"""
    active_profile = manager.get_active_profile()
    await telemetry.send_event(
        {
            "command": "graph status",
            **prepare_profile_for_telemetry(active_profile),
        }
    )

    pid_file = get_pid_file()

    if not pid_file.exists():
        print("No graph build running")
        sys.exit(1)

    try:
        pid = int(pid_file.read_text().strip())
        if is_process_running(pid):
            print(f"Graph build running (PID: {pid})")

            # Show log file info if it exists
            log_file = get_log_file()
            if log_file.exists():
                print("View logs: unpage graph logs --follow")
        else:
            cleanup_pid_file()
            print("No graph build running")
            sys.exit(1)
    except ValueError:
        cleanup_pid_file()
        print("No graph build running")
        sys.exit(1)



import asyncio

from pydantic import AwareDatetime

from unpage.models import Observation


class HasMetrics:
    """Capability for nodes that can retrieve metrics."""

    async def list_available_metrics(self) -> list[str]:
        """List the names of all available metrics for a given entity."""
        raise NotImplementedError

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        """Retrieve the data for a specific metric for a given entity."""
        raise NotImplementedError

    async def get_metrics(
        self,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
        metric_names: list[str] | None = None,
    ) -> list[Observation] | str:
        """Retrieve the data for multiple metrics for a given entity.

        If no metric names are provided, all available metrics will be retrieved.

        Args:
            node: The node for which to retrieve metrics.
            time_range_start: The start time of the time range to retrieve metrics for.
            time_range_end: The end time of the time range to retrieve metrics for.
            metric_names: The names of the metrics to retrieve.

        Returns:
            A list of metrics.
        """
        metric_names = metric_names or await self.list_available_metrics()
        metric_results = await asyncio.gather(
            *[self.get_metric(m, time_range_start, time_range_end) for m in metric_names],
            return_exceptions=True,
        )

        observations = []
        for result in metric_results:
            if isinstance(result, list):
                observations.extend(result)
            elif isinstance(result, Observation):
                observations.append(result)

        return observations



import warnings
from collections.abc import AsyncGenerator
from typing import Any, Literal

import anyio
import boto3.session
import rich
from aioboto3 import Session
from pydantic import BaseModel, Field, ValidationError
from pydantic_core import to_jsonable_python

from unpage.config import PluginSettings
from unpage.knowledge import Graph
from unpage.plugins import Plugin
from unpage.plugins.aptible.nodes.aptible_aws_instance import AptibleAwsInstance
from unpage.plugins.aws.nodes.aws_alb_target_group import AwsAlbTargetGroup
from unpage.plugins.aws.nodes.aws_application_load_balancer import (
    AwsApplicationLoadBalancer,
)
from unpage.plugins.aws.nodes.aws_classic_load_balancer import AwsClassicLoadBalancer
from unpage.plugins.aws.nodes.aws_ebs_volume import AwsEbsVolume
from unpage.plugins.aws.nodes.aws_ec2_instance import AwsEc2Instance
from unpage.plugins.aws.nodes.aws_rds_database import AwsRdsDatabase
from unpage.plugins.aws.nodes.aws_s3_bucket import AwsS3Bucket
from unpage.plugins.aws.nodes.base import DEFAULT_AWS_ACCOUNT_NAME, AwsAccount
from unpage.plugins.aws.utils import (
    ensure_aws_session,
    list_accessible_regions_for_service,
    swallow_boto_client_access_errors,
)
from unpage.plugins.mixins import KnowledgeGraphMixin, McpServerMixin, tool
from unpage.utils import Choice, classproperty, confirm, print, select

warnings.filterwarnings(
    "ignore",
    category=DeprecationWarning,
    message="datetime.datetime.utcnow() is deprecated and scheduled for removal in a future version. Use timezone-aware objects to represent datetimes in UTC: datetime.datetime.now(datetime.UTC).*",
)


class AwsPluginSettings(BaseModel):
    accounts: dict[str, AwsAccount] = Field(
        default_factory=lambda: {DEFAULT_AWS_ACCOUNT_NAME: AwsAccount()}
    )

    @property
    def account(self) -> AwsAccount:
        return next(iter(self.accounts.values()))


class AwsPlugin(Plugin, KnowledgeGraphMixin, McpServerMixin):
    aws_settings: AwsPluginSettings

    def __init__(
        self, *args: Any, aws_settings: AwsPluginSettings | None = None, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.aws_settings = aws_settings if aws_settings else AwsPluginSettings()

    def init_plugin(self) -> None:
        aws_accounts = self._settings.get("accounts")
        if not aws_accounts:
            self.aws_settings = AwsPluginSettings()
            return
        if not isinstance(aws_accounts, dict):
            raise ValueError("aws accounts must be a dictionary in config.yaml")
        if len(aws_accounts) != 1:
            raise ValueError(
                "More than one AWS account configured in config.yaml; we only support one AWS account at this time. Please let us know if you have more than one aws account."
            )
        for account_name, account_settings in aws_accounts.items():
            try:
                self.aws_settings = AwsPluginSettings(
                    accounts={
                        account_name: AwsAccount(
                            **{"name": account_name, **to_jsonable_python(account_settings)}
                        )
                    }
                )
            except ValidationError as ex:
                raise ValueError(
                    f"Invalid AWS account settings for aws account '{account_name}'. Review your config.yaml. {account_settings=}; error={ex!s}"
                ) from ex

    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        await ensure_aws_session(self.session)

    @classproperty
    def default_plugin_settings(cls) -> PluginSettings:
        return AwsPluginSettings().model_dump()

    async def interactive_configure(self) -> PluginSettings:
        rich.print("> The AWS plugin will add resources from AWS to your infra knowledge graph")
        rich.print(
            "> You can optionally set an AWS profile from your system that Unpage will use to authenticate AWS requests"
        )
        rich.print(
            "> Specifying no profile will fallback to using the standard AWS credential providers"
        )
        rich.print(
            "> Read more about the standard credential providers at: https://docs.aws.amazon.com/sdkref/latest/guide/standardized-credentials.html"
        )
        rich.print("")
        avail_aws_profiles = boto3.session.Session().available_profiles
        selected_aws_profile: list[str] = []
        if avail_aws_profiles and await confirm(
            "Would you like to select an AWS profile for Unpage to use when querying your AWS infrastructure?",
            default=True,
        ):
            enable_search = len(avail_aws_profiles) > 10
            choices = [
                Choice(p, checked=(p == self.aws_settings.account.profile))
                for p in avail_aws_profiles
            ]
            checked_choices = [c for c in choices if c.checked]
            default_choice = choices[0] if not checked_choices else checked_choices[0]
            selected_aws_profile = [
                await select(
                    "Which AWS profile would you like to use with the Unpage MCP Server?",
                    choices=choices,
                    default=default_choice,
                    use_search_filter=enable_search,
                    use_jk_keys=not enable_search,
                )
            ]
        settings = AwsPluginSettings(
            accounts={
                profile_name: AwsAccount(
                    name=profile_name,
                    profile=profile_name,
                )
                for profile_name in selected_aws_profile
            }
        )
        return settings.model_dump()

    @property
    def session(self) -> Session:
        if not hasattr(self, "_session"):
            self._session = self.aws_settings.account.session
        return self._session

    async def populate_graph(self, graph: Graph) -> None:
        await ensure_aws_session(self.session)
        async with anyio.create_task_group() as tg:
            tg.start_soon(self.populate_rds_databases, graph)
            tg.start_soon(self.populate_ec2_instances, graph)
            tg.start_soon(self.populate_classic_load_balancers, graph)
            tg.start_soon(self.populate_application_load_balancers, graph)
            tg.start_soon(self.populate_alb_target_groups, graph)
            tg.start_soon(self.populate_ebs_volumes, graph)
            tg.start_soon(self.populate_s3_buckets, graph)

    async def populate_rds_databases(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "rds"):
                tg.start_soon(self._populate_rds_databases_in_region, graph, region)

    async def _populate_rds_databases_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating RDS databases from {region}")

        rds_database_count = 0
        async for page in self._paginate("rds", "describe_db_instances", region):
            for instance in page["DBInstances"]:
                await graph.add_node(
                    AwsRdsDatabase(
                        node_id=instance["DBInstanceArn"],
                        raw_data=instance,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                rds_database_count += 1

        print(f"Initialized {rds_database_count} RDS databases for {region}")

    async def populate_ec2_instances(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "ec2"):
                tg.start_soon(self._populate_ec2_instances_in_region, graph, region)

    async def _populate_ec2_instances_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating EC2 instances from {region}")

        ec2_instance_count = 0
        async for page in self._paginate("ec2", "describe_instances", region):
            for reservation in page["Reservations"]:
                for instance in reservation["Instances"]:
                    await graph.add_node(
                        AwsEc2Instance(
                            node_id=instance["InstanceId"],
                            raw_data=instance,
                            _graph=graph,
                            aws_account=self.aws_settings.account,
                        )
                    )
                    ec2_instance_count += 1

        print(f"Initialized {ec2_instance_count} EC2 instances for {region}")

    async def populate_classic_load_balancers(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elb"):
                tg.start_soon(self._populate_classic_load_balancers_in_region, graph, region)

    async def _populate_classic_load_balancers_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating classic load balancers from {region}")

        elb_count = 0
        async for page in self._paginate("elb", "describe_load_balancers", region):
            for balancer in page["LoadBalancerDescriptions"]:
                await graph.add_node(
                    AwsClassicLoadBalancer(
                        node_id=balancer["LoadBalancerName"],
                        raw_data=balancer,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                elb_count += 1

        print(f"Initialized {elb_count} classic load balancers for {region}")

    async def populate_application_load_balancers(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elbv2"):
                tg.start_soon(self._populate_application_load_balancers_in_region, graph, region)

    async def _populate_application_load_balancers_in_region(
        self, graph: Graph, region: str
    ) -> None:
        print(f"Populating application load balancers from {region}")

        alb_count = 0
        async for page in self._paginate("elbv2", "describe_load_balancers", region):
            for balancer in page["LoadBalancers"]:
                await graph.add_node(
                    AwsApplicationLoadBalancer(
                        node_id=balancer["LoadBalancerArn"],
                        raw_data=balancer,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                alb_count += 1

        print(f"Initialized {alb_count} application load balancers for {region}")

    async def populate_alb_target_groups(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "elbv2"):
                tg.start_soon(self._populate_alb_target_groups_in_region, graph, region)

    async def _populate_alb_target_groups_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating ALB target groups from {region}")

        alb_target_group_count = 0
        async for page in self._paginate("elbv2", "describe_target_groups", region):
            for target_group in page["TargetGroups"]:
                await graph.add_node(
                    AwsAlbTargetGroup(
                        node_id=target_group["TargetGroupArn"],
                        raw_data=target_group,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                alb_target_group_count += 1

        print(f"Initialized {alb_target_group_count} ALB target groups for {region}")

    async def populate_ebs_volumes(self, graph: Graph) -> None:
        async with anyio.create_task_group() as tg:
            for region in await list_accessible_regions_for_service(self.session, "ec2"):
                tg.start_soon(self._populate_ebs_volumes_in_region, graph, region)

    async def _populate_ebs_volumes_in_region(self, graph: Graph, region: str) -> None:
        print(f"Populating EBS volumes from {region}")

        ebs_volume_count = 0
        async for page in self._paginate("ec2", "describe_volumes", region):
            for volume in page["Volumes"]:
                await graph.add_node(
                    AwsEbsVolume(
                        node_id=volume["VolumeId"],
                        raw_data=volume,
                        _graph=graph,
                        aws_account=self.aws_settings.account,
                    )
                )
                ebs_volume_count += 1

        print(f"Initialized {ebs_volume_count} EBS volumes for {region}")

    async def populate_s3_buckets(self, graph: Graph) -> None:
        # S3 buckets are global resources, not regional like other AWS services
        # We only need to call list_buckets once, not per region
        print("Populating S3 buckets")

        s3_bucket_count = 0
        async with (
            swallow_boto_client_access_errors(service_name="s3", region="us-east-1"),
            self.session.client("s3", region_name="us-east-1") as client,
        ):
            paginator = client.get_paginator("list_buckets")
            async for page in paginator.paginate():
                for bucket in page["Buckets"]:
                    await graph.add_node(
                        AwsS3Bucket(
                            node_id=bucket["Name"],
                            raw_data=bucket,
                            _graph=graph,
                            aws_account=self.aws_settings.account,
                        )
                    )
                    s3_bucket_count += 1

        print(f"Initialized {s3_bucket_count} S3 buckets")

    @tool()
    async def get_realtime_instance_status(self, instance_id: str, region: str) -> dict | str:
        """
        Get real-time status information for an EC2 instance directly from AWS API.

        Args:
            instance_id: Optional AWS EC2 instance ID
            region: AWS region where the instance is located

        Returns:
            dict containing current instance state and status details
        """
        async with (
            swallow_boto_client_access_errors(service_name="ec2", region=region),
            self.session.client("ec2", region_name=region) as client,
        ):
            try:
                response = await client.describe_instance_status(InstanceIds=[instance_id])
                if not response["InstanceStatuses"]:
                    return f"No instance found with ID '{instance_id}' in region '{region}'"
                return response["InstanceStatuses"][0]
            except Exception as e:
                return f"Error retrieving instance status: {e!s}"

    @tool()
    async def get_realtime_instance_status_by_node(self, node_id: str) -> dict | str:
        """
        Get real-time status information for an EC2 instance node directly from AWS API.

        Args:
            node_id: node ID from the knowledge graph

        Returns:
            dict containing current instance state and status details
        """
        node = await self.context.graph.get_node_safe(node_id)
        if not node:
            return f"Resource with node ID '{node_id}' not found"
        if isinstance(node, AwsEc2Instance):
            return await self.get_realtime_instance_status(
                node.raw_data["InstanceId"], node.raw_data["Placement"]["AvailabilityZone"][:-1]
            )
        elif isinstance(node, AptibleAwsInstance) and "instance_id" in node.raw_data:
            return await self.get_realtime_instance_status(
                node.raw_data["instance_id"], node.raw_data["availability_zone"][:-1]
            )
        else:
            return f"Node {node_id} is not an EC2 instance or does not have an instance ID"

    async def _paginate(
        self,
        service_name: Literal["ec2", "elb", "elbv2", "rds"],
        action: str,
        region: str,
    ) -> AsyncGenerator[dict, None]:
        async with (
            swallow_boto_client_access_errors(service_name=service_name, region=region),
            self.session.client(service_name, region_name=region) as client,
        ):
            paginator = client.get_paginator(action)
            async for page in paginator.paginate():
                yield page



import weakref
from collections.abc import AsyncGenerator, Callable

from datadog_api_client import AsyncApiClient, Configuration
from datadog_api_client.v2.api.logs_api import LogsApi
from datadog_api_client.v2.api.software_catalog_api import SoftwareCatalogApi
from datadog_api_client.v2.api.teams_api import TeamsApi
from datadog_api_client.v2.model.entity_data import EntityData
from datadog_api_client.v2.model.include_type import IncludeType
from datadog_api_client.v2.model.list_teams_include import ListTeamsInclude
from datadog_api_client.v2.model.team import Team
from pydantic import AwareDatetime

from unpage.plugins.datadog.models import DatadogLogEvent


class DatadogClient:
    """A thin wrapper around the Datadog API client."""

    def __init__(self, api_key: str, application_key: str) -> None:
        self.client = AsyncApiClient(
            configuration=Configuration(
                api_key={
                    "apiKeyAuth": api_key,
                    "appKeyAuth": application_key,
                }
            )
        )
        # Register a finalizer to close the client when this object is garbage collected
        self._finalizer = weakref.finalize(self, lambda: self.client.close())

    async def iter_teams(self) -> AsyncGenerator[Team, None]:
        teams = TeamsApi(self.client)
        async for team in teams.list_teams_with_pagination(
            include=[ListTeamsInclude.USER_TEAM_PERMISSIONS]
        ):
            yield team

    async def iter_software_catalog_entities(
        self,
    ) -> AsyncGenerator[EntityData, None]:
        software_catalog = SoftwareCatalogApi(self.client)
        async for entity in software_catalog.list_catalog_entity_with_pagination(
            include=IncludeType.RELATION
        ):
            yield entity

    async def search_logs(
        self,
        query: str,
        min_time: AwareDatetime,
        max_time: AwareDatetime,
        continue_search: Callable[[AwareDatetime | None], bool] | None = None,
    ) -> AsyncGenerator[DatadogLogEvent, None]:
        """Search logs in Datadog"""
        logs = LogsApi(self.client)
        async for log in logs.list_logs_get_with_pagination(
            filter_query=query,
            filter_from=min_time,
            filter_to=max_time,
        ):
            event = DatadogLogEvent(**log.to_dict())
            yield event
            if continue_search is not None and not continue_search(event.attributes.timestamp):
                break



import json
from typing import cast

from pydantic import AwareDatetime

from unpage.knowledge import NODE_REGISTRY, HasMetrics
from unpage.models import Observation
from unpage.plugins.base import Plugin
from unpage.plugins.mixins import McpServerMixin, tool


class MetricsPlugin(Plugin, McpServerMixin):
    @tool()
    async def list_node_types_that_support_metrics(self) -> list[str]:
        """List the types of nodes that support metrics retrieval.

        This is useful for getting a list of node types that you can use in
        metrics_get_metrics_for_resource.

        Use this before attempting to get metrics for a resource.
        """
        return [
            f"{node_cls._node_source}:{node_cls._node_type}"
            for node_cls in NODE_REGISTRY.values()
            if issubclass(node_cls, HasMetrics)
        ]

    @tool()
    async def list_available_metrics_for_node(self, node_id: str) -> list[str] | str:
        """List the available metrics for a node."""
        node = await self.context.graph.get_node_safe(node_id)
        if not node:
            return f"Resource with node ID '{node_id}' not found"

        if not isinstance(node, HasMetrics):
            return f"{node.node_type} nodes do not support metrics"

        return await node.list_available_metrics()

    @tool()
    async def get_metrics_for_node(
        self,
        node_id: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
        metric_names: list[str] | str | None = None,
    ) -> list[Observation] | str:
        """Get metrics for a resource using its node ID.

        A good starting point for the length of the time range is an hour.

        You can get valid metric names for a node using the
        list_available_metrics_for_node tool.

        Metric names should be provided as a JSON array of strings.
        """
        node = await self.context.graph.get_node_safe(node_id)
        if not node:
            return f"Resource with node ID '{node_id}' not found"

        if not isinstance(node, HasMetrics):
            return f"{node.node_type} nodes do not support metrics"

        metric_names = metric_names if metric_names else await node.list_available_metrics()

        # HACK: Some LLM clients send the metric names as a JSON string, which
        # we need to parse into a list.
        if isinstance(metric_names, str):
            try:
                metric_names = json.loads(metric_names)
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"metric_names should be a valid JSON array, got {metric_names}"
                ) from e

        observations = await node.get_metrics(
            metric_names=cast("list[str]", metric_names),
            time_range_start=time_range_start,
            time_range_end=time_range_end,
        )

        return observations or "No metrics found. Try a longer time range."



import re
from typing import Self

from pydantic import BaseModel

_arn_base_pattern = re.compile(
    r"^arn:"
    r"(?P<partition>.+?):"
    r"(?P<service>.+?):"
    r"(?P<region>.*?):"
    r"(?P<account>.*?):"
    r"(?P<rest>.*)$"
)


class AwsArn(BaseModel):
    partition: str
    service: str
    region: str | None = None
    account_id: str | None = None
    rest: str | None = None

    @classmethod
    def parse(cls, arn: str) -> Self:
        base_match = _arn_base_pattern.match(arn)
        if not base_match:
            raise ValueError(f"Invalid ARN format: {arn}")
        return cls(
            partition=base_match.group("partition"),
            service=base_match.group("service"),
            region=base_match.group("region"),
            account_id=base_match.group("account"),
            rest=base_match.group("rest"),
        )



import re

from unpage.knowledge import Node
from unpage.utils import camel_to_snake


class DatadogNode(Node):
    """A base class for Datadog nodes."""


class DatadogEntityNode(DatadogNode):
    """A base class for Datadog entity nodes."""

    _RELATED_ENTITY_REGEX = re.compile(
        r"^(?P<self_kind>[^:]+):(?P<self_namespace>[^/]+)/(?P<self_name>[^:]+):"
        r"RelationType(?P<relation>[^:]+):"
        r"(?P<other_kind>[^:]+):(?P<other_namespace>[^/]+)/(?P<other_name>[^:]+)$"
    )

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            *await self._get_related_entity_reference_identifiers(),
        ]

    async def _get_related_entity_reference_identifiers(
        self,
    ) -> list[tuple[str, str]]:
        """Get the reference identifiers for the related entities."""
        referenced_entity_identifiers: list[tuple[str, str]] = []

        related_entities = (
            self.raw_data.get("relationships", {}).get("related_entities", {}).get("data", [])
        )

        for related_entity in related_entities:
            parsed_id = self._RELATED_ENTITY_REGEX.match(related_entity["id"])

            if parsed_id is None:
                continue

            self_kind = parsed_id.group("self_kind")
            self_namespace = parsed_id.group("self_namespace")
            self_name = parsed_id.group("self_name")
            relation = camel_to_snake(parsed_id.group("relation"))
            other_kind = parsed_id.group("other_kind")
            other_namespace = parsed_id.group("other_namespace")
            other_name = parsed_id.group("other_name")

            print(
                f"{self_kind}:{self_namespace}/{self_name} --[{relation}]--> {other_kind}:{other_namespace}/{other_name}"
            )

            referenced_entity_identifiers.append(
                (
                    f"{other_kind}:{other_namespace}/{other_name}",
                    relation,
                )
            )

        return referenced_entity_identifiers



from typing import Any

from fastmcp import FastMCP
from fastmcp.contrib.mcp_mixin import (
    MCPMixin,
    mcp_prompt,
    mcp_resource,
    mcp_tool,
)

from unpage.plugins import PluginCapability


class McpServerMixin(MCPMixin, PluginCapability):
    """Capability for registering tools, resources, and prompts with the MCP server."""

    def register_all(self, mcp_server: FastMCP[Any], *args: Any, **kwargs: Any) -> None:
        # Register any existing resources, as usual.
        super().register_all(mcp_server, *args, **kwargs)

        # Mount the MCP sub-server
        mcp_server.mount(self.get_mcp_server())

    def get_mcp_server(self) -> FastMCP[Any]:
        return FastMCP[Any](self.name)


prompt = mcp_prompt
resource = mcp_resource
tool = mcp_tool



import ast
import operator
from datetime import UTC, datetime
from zoneinfo import ZoneInfo

from pydantic import AwareDatetime

from unpage.plugins.base import Plugin
from unpage.plugins.mixins.mcp import McpServerMixin, tool


class CorePlugin(Plugin, McpServerMixin):
    @tool()
    def current_datetime(self) -> AwareDatetime:
        """Return the current date and time in UTC."""
        return datetime.now(UTC)

    @tool()
    def convert_to_timezone(self, datetime: AwareDatetime, timezone: str) -> AwareDatetime:
        """Convert a datetime to a specific timezone."""
        return datetime.astimezone(ZoneInfo(timezone))

    @tool()
    def calculate(self, expression: str) -> str:
        """Evaluate a mathematical expression."""
        allowed_operators = {
            ast.Add: operator.add,
            ast.Sub: operator.sub,
            ast.Mult: operator.mul,
            ast.Div: operator.truediv,
            ast.FloorDiv: operator.floordiv,
            ast.Mod: operator.mod,
            ast.Pow: operator.pow,
            ast.USub: operator.neg,
        }

        def eval_expr(node: ast.AST) -> float:
            if isinstance(node, ast.Constant):
                if isinstance(node.value, str | int | float):
                    return float(node.value)
                raise ValueError(f"Unsupported constant type: {type(node.value)}")
            elif isinstance(node, ast.BinOp):
                left = eval_expr(node.left)
                right = eval_expr(node.right)
                op_type = type(node.op)
                if op_type in allowed_operators:
                    return allowed_operators[op_type](left, right)
            elif isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.USub):
                return eval_expr(ast.UnaryOp(op=ast.USub(), operand=node.operand))
            raise ValueError(f"Unsupported operation: {ast.dump(node)}")

        return str(eval_expr(ast.parse(expression, mode="eval").body))



import sys
from typing import Annotated

import rich
from cyclopts import App, Group, Parameter

from unpage.cli.agent._app import agent_app
from unpage.cli.graph._app import graph_app
from unpage.cli.mcp._app import mcp_app
from unpage.cli.mlflow._app import mlflow_app
from unpage.cli.profile._app import profile_app
from unpage.config import manager
from unpage.warnings import filter_all_warnings

filter_all_warnings()

app = App(
    default_parameter=Parameter(
        # Disable automatic creation of "negative" options (e.g. --no-foo)
        negative=()
    )
)

app.command(agent_app, name="agent")
app.command(graph_app, name="graph")
app.command(mcp_app, name="mcp")
app.command(mlflow_app, name="mlflow")
app.command(profile_app, name="profile")


@app.meta.default
def _app_launcher(
    *tokens: Annotated[str, Parameter(show=False, allow_leading_hyphen=True)],
    profile: str = manager.get_active_profile(),
) -> None:
    # Let the user know when they're overriding the active profile.
    if profile != manager.get_active_profile():
        rich.print(f"[blue]Active profile: {profile}[/blue]", file=sys.stderr)

    with manager.active_profile(profile):
        app(tokens)


# Rename the meta's "Parameter" -> "Session Parameters".
# Set sort_key so it will be drawn higher up the help-page.
app.meta.group_parameters = Group("Session Parameters", sort_key=0)



import rich

from unpage.cli.profile._app import profile_app
from unpage.config import manager


@profile_app.command(name="list")
def list_profiles() -> None:
    """List all available profiles."""
    profiles = manager.list_profiles()
    active_profile = manager.get_active_profile()

    rich.print("Available profiles:")
    for profile in profiles:
        if profile == active_profile:
            rich.print(f"  [green]* {profile}[/green] (active)")
        else:
            rich.print(f"    {profile}")



from pathlib import Path

from pydantic import BaseModel, Field
from pydantic_yaml import parse_yaml_file_as


class Prompt(BaseModel):
    trigger: str = Field(description="A description of when this prompt should be used")
    runbook: str = Field(description="The runbook for this prompt")


PROMPTS_DIR = Path(__file__).parent
PROMPT_FILES = list(PROMPTS_DIR.glob("**/*.yaml"))
ALL_PROMPTS = {f.stem: parse_yaml_file_as(Prompt, f) for f in PROMPT_FILES}



import sys
from typing import TYPE_CHECKING, cast

from rich import print

from unpage.agent.analysis import AnalysisAgent
from unpage.agent.utils import load_agent
from unpage.cli.agent._app import agent_app
from unpage.config import manager
from unpage.plugins.base import PluginManager
from unpage.telemetry import client as telemetry
from unpage.telemetry import hash_value, prepare_profile_for_telemetry

if TYPE_CHECKING:
    from unpage.plugins.pagerduty.plugin import PagerDutyPlugin
    from unpage.plugins.rootly.plugin import RootlyPlugin


@agent_app.command
async def run(
    agent_name: str,
    payload: str | None = None,
    /,
    *,
    pagerduty_incident: str | None = None,
    rootly_incident: str | None = None,
    debug: bool = False,
    use_test_payload: str | None = None,
) -> None:
    """Run an agent with the provided payload and print the analysis.

    A payload can be passed as an argument or piped to stdin.

    Parameters
    ----------
    agent_name
        The name of the agent to run
    payload
        The alert payload to analyze. Alternatively, you can pipe the payload to stdin.
    pagerduty_incident
        PagerDuty incident ID or URL to use instead of payload or stdin
    rootly_incident
        Rootly incident ID or URL to use instead of payload or stdin
    debug
        Enable debug mode to print the history of the agent
    use_test_payload
        Use a predefined test payload defined in your agent yaml file (e.g. "pagerduty_incident" or "build_failure")
    """
    await telemetry.send_event(
        {
            "command": "agent run",
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "agent_name_sha256": hash_value(agent_name),
            "debug": debug,
            "has_payload": payload is not None,
            "has_pagerduty_incident": bool(pagerduty_incident),
            "has_rootly_incident": bool(rootly_incident),
        }
    )
    plugin_manager = PluginManager(manager.get_active_profile_config())
    data = ""
    if (
        sum(
            [
                bool(pagerduty_incident),
                bool(rootly_incident),
                bool(use_test_payload),
                payload is not None or not sys.stdin.isatty(),
            ]
        )
        >= 2
    ):
        print(
            "[red]Cannot pass --pagerduty-incident, --rootly-incident, or --use-test-payload with --payload or stdin.[/red]"
        )
        sys.exit(1)

    if pagerduty_incident:
        incident_id = pagerduty_incident
        if "/" in pagerduty_incident:
            incident_id = [x for x in pagerduty_incident.split("/") if x][-1]
        pd = cast("PagerDutyPlugin", plugin_manager.get_plugin("pagerduty"))
        incident = await pd.get_incident_by_id(incident_id)
        data = incident.model_dump_json()

    if rootly_incident:
        incident_id = rootly_incident
        if "/" in rootly_incident:
            incident_id = [x for x in rootly_incident.split("/") if x][-1]
        rootly = cast("RootlyPlugin", plugin_manager.get_plugin("rootly"))
        incident = await rootly.get_incident_by_id(incident_id)
        data = incident.model_dump_json()

    if use_test_payload:
        agent = load_agent(agent_name)
        if not agent.test_payloads or use_test_payload not in agent.test_payloads:
            print(
                f"[red]Test payload {use_test_payload!r} not found in agent {agent_name!r}.[/red]"
            )
            if agent.test_payloads:
                print(
                    f"[bold]Available test payloads: {', '.join(agent.test_payloads.keys())}[/bold]"
                )
            sys.exit(1)
        data = agent.test_payloads[use_test_payload]

    # Read data from stdin if it's being piped to us.
    if not data and not sys.stdin.isatty():
        if payload is not None:
            print("[red]Cannot pass a payload argument when piping data to stdin.[/red]")
            sys.exit(1)
        data = sys.stdin.read().strip()
    elif not data:
        # Otherwise, use the payload argument.
        data = payload

    if not data:
        print("[red]No payload provided.[/red]")
        print(
            "[bold]Pass an alert payload as an argument or pipe the payload data to stdin.[/bold]"
        )
        sys.exit(1)

    # Get the config directory and load the specific agent
    try:
        agent = load_agent(agent_name)
    except FileNotFoundError as ex:
        print(f"[red]Agent {agent_name!r} not found at {str(ex.filename)!r}[/red]")
        print(f"[bold]Use 'unpage agent create {agent_name!r}' to create a new agent.[/bold]")
        sys.exit(1)

    # Run the analysis with the specific agent
    analysis_agent = AnalysisAgent()
    try:
        result = await analysis_agent.acall(payload=data, agent=agent)
        print(result)
    except Exception as ex:
        print(f"[red]Analysis failed:[/red] {ex}")
        sys.exit(1)
    finally:
        if debug:
            print("\n\n===== DEBUG OUTPUT =====\n")
            analysis_agent.inspect_history(n=1000)
            print("\n===== END DEBUG OUTPUT =====\n\n")



"""Configuration management for unpage CLI."""

import shutil
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Self

import yaml
from expandvars import expandvars
from pydantic import BaseModel, Field, model_validator
from pydantic_settings import BaseSettings


def yaml_to_model[T: BaseModel](
    model_class: type[T],
    yaml_content: str | Path,
    **overrides: Any,
) -> T:
    """Load YAML content into a Pydantic model with optional field overrides.

    Args:
        model_class: The Pydantic model class to create
        yaml_content: YAML content as string or Path to YAML file
        **overrides: Additional keyword arguments to override/merge with YAML values

    Returns:
        Instance of the model class with data from YAML and overrides

    Raises:
        FileNotFoundError: If yaml_content is a Path that doesn't exist
        yaml.YAMLError: If YAML content is invalid
        pydantic.ValidationError: If the resulting data doesn't match the model
    """
    # Load YAML content
    content = yaml_content.read_text() if isinstance(yaml_content, Path) else yaml_content

    # Parse YAML
    yaml_data = yaml.safe_load(content) or {} if content.strip() else {}

    # Ensure data is a dictionary
    if not isinstance(yaml_data, dict):
        raise ValueError(f"YAML content must be a dictionary, got {type(yaml_data).__name__}")

    # Create and return model instance
    return model_class(
        **{
            **yaml_data,
            # Merge overrides into data (overrides take precedence)
            **overrides,
        }
    )


class Environment(BaseSettings):
    """Environment variables for unpage configuration."""

    UNPAGE_CONFIG_ROOT: Path | None = None

    def get_config_root(self) -> Path:
        """Get the configuration directory, with fallback to default."""
        return self.UNPAGE_CONFIG_ROOT or (Path.home() / ".unpage")


# Global environment instance
env = Environment()


type PluginSettings = dict[str, Any]
"""type alias to dict[str, Any] that contains settings for the plugin"""


class PluginConfig(BaseModel):
    enabled: bool = True
    settings: PluginSettings = Field(default_factory=dict)

    def __hash__(self) -> int:
        return hash(self.model_dump_json())


class EnvironmentVariablesMixin(BaseModel):
    """Model to recursively expand environment variables in all fields."""

    @model_validator(mode="before")
    @classmethod
    def expand_env_vars(cls, data: Any) -> Any:  # noqa: ANN401
        """Recursively expand environment variables in all string fields."""
        if not isinstance(data, dict):
            return data

        def _expand(value: Any) -> Any:  # noqa: ANN401
            if isinstance(value, str):
                return expandvars(value, nounset=True)
            elif isinstance(value, dict):
                return {k: _expand(v) for k, v in value.items()}
            elif isinstance(value, list):
                return [_expand(item) for item in value]
            else:
                return value

        return _expand(data)


class Config(EnvironmentVariablesMixin, BaseModel):
    """Configuration model for unpage profiles."""

    plugins: dict[str, PluginConfig] = Field(default_factory=dict)
    profile: str = Field(default="default", exclude=True)
    file_path: Path = Field(exclude=True)

    telemetry_enabled: bool = True

    def save(self) -> None:
        """Save this config to its file path."""
        if not self.file_path:
            raise ValueError("Cannot save config: file_path not set")
        self.file_path.parent.mkdir(parents=True, exist_ok=True)
        with self.file_path.open("w") as f:
            yaml.dump(self.model_dump(), f, default_flow_style=False)

    def merge_plugins(self, other_plugins: dict[str, PluginConfig]) -> Self:
        """Merge another plugins dict into this config's plugins, returning a new Config instance.

        Args:
            other_plugins: Another plugins dict to mergee, may be None in which case self is returned

        Returns:
            Config new instance with merged values
        """
        if not other_plugins:
            return self

        return self.__class__(
            **{
                "file_path": self.file_path,
                "profile": self.profile,
                **self.model_dump(),
                "plugins": {**self.plugins, **other_plugins},
            }
        )


class ConfigManager:
    """Manages user configuration and profiles for unpage."""

    def __init__(self, config_root: Path | None = None) -> None:
        """Initialize configuration manager.

        Args:
            config_root: Custom config directory. Defaults to UNPAGE_CONFIG_ROOT or ~/.unpage
        """
        self.config_root = config_root or env.get_config_root()
        self.profiles_dir = self.config_root / "profiles"
        self.active_profile_file = self.config_root / ".profile"
        self.active_profile_override = None
        self._ensure_config_structure()

    def _ensure_config_structure(self) -> None:
        """Ensure the configuration directory structure exists."""
        self.config_root.mkdir(exist_ok=True)
        self.profiles_dir.mkdir(exist_ok=True)

        # Create default profile if it doesn't exist
        default_profile = self.profiles_dir / "default"
        if not default_profile.exists():
            self.create_profile("default")

        # Set default profile as current if no profile is set
        if not self.active_profile_file.exists():
            self.set_active_profile("default")

    def get_empty_config(self, profile: str, **overrides: Any) -> Config:
        """Get an empty config object."""
        # Need to import here to avoid circular import
        from unpage.plugins.base import REGISTRY

        return Config(
            profile=profile,
            file_path=self.profiles_dir / profile / "config.yaml",
            **{
                "plugins": {
                    p: PluginConfig(
                        enabled=plugin_cls.default_enabled,
                        settings=plugin_cls.default_plugin_settings,
                    )
                    for p, plugin_cls in REGISTRY.items()
                    if plugin_cls.default_enabled
                },
                **overrides,
            },
        )

    def create_profile(self, name: str) -> Path:
        """Create a new profile with an empty config.yaml.

        Args:
            name: Name of the profile to create

        Returns:
            Path to the created profile directory

        Raises:
            FileExistsError: If profile already exists
        """
        profile_dir = self.profiles_dir / name

        if profile_dir.exists():
            raise FileExistsError(f"Profile {name!r} already exists")

        profile_dir.mkdir(parents=True, exist_ok=True)

        # Create empty config file using Config model
        empty_config = Config(profile=name, file_path=profile_dir / "config.yaml")
        empty_config.save()

        return profile_dir

    def list_profiles(self) -> list[str]:
        """List all available profiles.

        Returns:
            List of profile names
        """
        if not self.profiles_dir.exists():
            return []

        return [
            p.name
            for p in self.profiles_dir.iterdir()
            if p.is_dir() and (p / "config.yaml").exists()
        ]

    def get_profile_directory(self, name: str) -> Path:
        """Get the directory for a specific profile.

        Args:
            name: Profile name

        Returns:
            Path to the profile directory
        """
        return self.profiles_dir / name

    def get_profile_config(self, name: str) -> Config:
        """Get configuration for a specific profile.

        Args:
            name: Profile name

        Returns:
            Configuration object

        Raises:
            FileNotFoundError: If profile doesn't exist
        """
        profile_dir = self.profiles_dir / name
        config_file = profile_dir / "config.yaml"

        if not config_file.exists():
            raise FileNotFoundError(f"Profile {name!r} not found")

        # Load config from YAML file with overrides
        return yaml_to_model(Config, config_file, profile=name, file_path=config_file)

    def get_active_profile(self) -> str:
        """Get the name of the currently active profile.

        Returns:
            Name of active profile
        """
        if self.active_profile_override:
            return self.active_profile_override

        if not self.active_profile_file.exists():
            return "default"

        return self.active_profile_file.read_text().strip()

    def set_active_profile(self, name: str) -> None:
        """Set the current active profile.

        Args:
            name: Profile name to set as active

        Raises:
            FileNotFoundError: If profile doesn't exist
        """
        profile_dir = self.profiles_dir / name
        if not profile_dir.exists():
            raise FileNotFoundError(f"Profile {name!r} not found")

        self.active_profile_file.write_text(name)

    @contextmanager
    def active_profile(self, name: str) -> Generator[None, None, None]:
        """Context manager for setting the active profile."""
        self.active_profile_override = name
        try:
            yield
        finally:
            self.active_profile_override = None

    def get_active_profile_directory(self) -> Path:
        """Get the directory for the currently active profile.

        Returns:
            Path to the active profile directory
        """
        active_profile = self.get_active_profile()
        return self.get_profile_directory(active_profile)

    def get_active_profile_config(self) -> Config:
        """Get configuration for the currently active profile.

        Returns:
            Configuration object for the active profile
        """
        active_profile = self.get_active_profile()
        return self.get_profile_config(active_profile)

    def delete_profile(self, name: str) -> None:
        """Delete a profile.

        Args:
            name: Profile name to delete
        """
        profile_dir = self.profiles_dir / name
        if not profile_dir.exists():
            raise FileNotFoundError(f"Profile {name!r} not found")

        shutil.rmtree(profile_dir)


# Global config manager instance
manager = ConfigManager()



import asyncio
import shutil
import sys

from unpage.cli.graph._app import graph_app
from unpage.cli.graph._background import get_log_file
from unpage.config import manager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@graph_app.command
async def logs(
    *,
    follow: bool = False,
) -> None:
    """View graph build logs

    Parameters
    ----------
    follow
        Follow log output
    """
    active_profile = manager.get_active_profile()
    await telemetry.send_event(
        {
            "command": "graph logs",
            **prepare_profile_for_telemetry(active_profile),
            "follow": follow,
        }
    )
    log_file = get_log_file()
    tail_cmd = shutil.which("tail")
    if not tail_cmd:
        print("'tail' command not found. Please install it.")
        sys.exit(1)

    if not log_file.exists():
        print("No log file found")
        print(f"Expected location: {log_file}")
        sys.exit(1)

    if follow:
        print("Following logs (Ctrl+C to stop)")
        print(f"Log file: {log_file}")

        try:
            proc = await asyncio.create_subprocess_shell(f"{tail_cmd} -f {log_file!s}")
            await proc.wait()
        except KeyboardInterrupt:
            print("\nStopped following logs")
    else:
        print("Recent logs:")
        print(f"Log file: {log_file}")

        proc = await asyncio.create_subprocess_shell(f"{tail_cmd} -n 50 {log_file!s}")
        await proc.wait()



import json
import sys

from fastmcp import Client
from mcp.types import TextContent

from unpage.cli.mcp.tools._app import tools_app
from unpage.config import manager
from unpage.knowledge import Graph
from unpage.mcp import Context, build_mcp_server
from unpage.plugins import PluginManager
from unpage.telemetry import client as telemetry
from unpage.telemetry import prepare_profile_for_telemetry


@tools_app.command
async def call(
    tool: str,
    /,
    *,
    count_results: bool = False,
    count_level: int = 0,
    arguments: list[str] | None = None,
) -> None:
    """Call an MCP tool from the command line.

    Parameters
    ----------
    tool
        The tool name to call
    count_results
        Count the number of results returned by the tool instead of printing the results
        (works best with list or dict tool responses)
    count_level
        Count level for nested structures (0 or 1)
    arguments
        Arguments to pass to the tool
    """
    await telemetry.send_event(
        {
            "command": "mcp tools call",
            "tool": tool,
            **prepare_profile_for_telemetry(manager.get_active_profile()),
            "count_results": count_results,
            "count_level": count_level,
            "has_arguments": arguments is not None,
        }
    )
    config = manager.get_active_profile_config()
    plugins = PluginManager(config=config)
    context = Context(
        profile=manager.get_active_profile(),
        config=config,
        plugins=plugins,
        graph=Graph(manager.get_active_profile_directory() / "graph.json"),
    )
    mcp = await build_mcp_server(context)

    tools = await mcp.get_tools()
    if tool not in tools:
        print(f"Tool {tool} not found")
        sys.exit(1)

    tool_def = tools[tool]
    tool_args = {}
    if arguments:
        # Convert a list of arguments into a dict
        for k, v in zip(tool_def.parameters["properties"].keys(), arguments, strict=True):
            # Try decoding as JSON to handle complex arguments (lists, dicts, etc.)
            try:
                tool_args[k] = json.loads(v)
            except json.JSONDecodeError:
                tool_args[k] = v

    client = Client(mcp)
    async with client:
        result = await client.call_tool(tool, tool_args)
        try:
            content = next(r.text for r in result.content if isinstance(r, TextContent))
            if count_results and count_level == 0:
                print(
                    json.dumps(
                        {"count": len(json.loads(content)), "content_length": len(content)},
                        indent=2,
                    )
                )
            elif count_results and count_level == 1:
                print(
                    json.dumps(
                        {
                            k: (
                                {
                                    "count": len(v),
                                    "content_length": len(json.dumps(v)),
                                }
                                if isinstance(v, list | dict)
                                else v
                            )
                            for k, v in json.loads(content).items()
                        },
                        indent=2,
                    )
                )
            else:
                print(content)
        except StopIteration:
            print(f"{tool} returned no printable results")



from collections import defaultdict

import anyio
import questionary
import rich

from unpage.config import PluginSettings
from unpage.knowledge import Graph, Node
from unpage.knowledge.edges import Edge
from unpage.plugins import Plugin
from unpage.plugins.aptible.client import AptibleClient
from unpage.plugins.aptible.nodes.base import inflate_resource
from unpage.plugins.mixins import KnowledgeGraphMixin
from unpage.utils import print


class AptiblePlugin(Plugin, KnowledgeGraphMixin):
    """A plugin for the Aptible PaaS."""

    async def interactive_configure(self) -> PluginSettings:
        rich.print("> The Aptible plugin uses your local system authentication.")
        rich.print("> Ensure you are logged into your preferred Aptible organization with:")
        rich.print(">")
        rich.print(">   aptible login")
        rich.print(">")
        rich.print("> Ready to proceed?")
        rich.print("")
        await questionary.press_any_key_to_continue().unsafe_ask_async()
        rich.print("")
        return {}

    async def validate_plugin_config(self) -> None:
        await super().validate_plugin_config()
        async with AptibleClient() as client:
            await client.validate_auth()

    async def populate_graph(self, graph: Graph) -> None:
        """Initialize the graph with nodes and edges from the Aptible API."""
        self.client = AptibleClient()

        edge_count = 0
        seen_nodes = defaultdict(set)

        async with self.client.concurrent_paginator("/edges") as (
            edges,
            total_count,
        ):
            notify_threshold = max(int(total_count * 0.01), 1)

            async for edge in edges:
                source_node = inflate_resource(
                    edge["_embedded"]["source_resource"],
                    _graph=graph,
                )
                destination_node = inflate_resource(
                    edge["_embedded"]["destination_resource"],
                    _graph=graph,
                )

                await graph.add_edge(
                    Edge(
                        source_node=source_node,
                        destination_node=destination_node,
                        properties={
                            "relationship_type": edge["relationship_type"],
                        },
                    )
                )

                if source_node.nid not in seen_nodes[source_node.node_type]:
                    seen_nodes[source_node.node_type].add(source_node.nid)
                if destination_node.nid not in seen_nodes[destination_node.node_type]:
                    seen_nodes[destination_node.node_type].add(destination_node.nid)

                edge_count += 1

                if edge_count % notify_threshold == 0:
                    percent_complete = (edge_count / total_count) * 100
                    print(f"Edge loading: {edge_count}/{total_count} ({percent_complete:.2f}%)")

        # Update app configurations
        async def _update_app_configuration(node: Node) -> None:
            # If the app has no link to a current_configuration, skip it.
            if "current_configuration" not in node.raw_data["_links"]:
                return

            # If the app has no services, it won't have a configuration, so skip it.
            if not node.raw_data["_embedded"]["services"]:
                return

            print(f"Retrieving the current configuration for {node.nid}")
            response = await self.client.get(
                node.raw_data["_links"]["current_configuration"]["href"]
            )

            if response.status_code != 200:
                print(f"Failed to retrieve the current configuration for {node.nid}")
                return

            node.raw_data["_embedded"]["current_configuration"] = response.json()

        async with anyio.create_task_group() as tg:
            async for node in graph.iter_nodes():
                if node.node_type == "aptible_app":
                    tg.start_soon(_update_app_configuration, node)

        total_nodes = sum(len(seen_nodes[node_type]) for node_type in seen_nodes)

        print(f"Initialized {edge_count} edges between {total_nodes} nodes")
        for node_type, node_ids in seen_nodes.items():
            print(f"  - {node_type}: {len(node_ids)}")



from pydantic import AwareDatetime

from unpage.knowledge import HasMetrics
from unpage.models import Observation

from .base import AwsNode


class AwsClassicLoadBalancer(AwsNode, HasMetrics):
    async def get_identifiers(self) -> list[str | None]:
        return [
            *await super().get_identifiers(),
            self.raw_data["LoadBalancerName"],
            self.raw_data["DNSName"],
        ]

    async def get_reference_identifiers(
        self,
    ) -> list[str | None | tuple[str | None, str]]:
        return [
            *await super().get_reference_identifiers(),
            self.raw_data.get("VpcId"),
            *self.raw_data["Subnets"],
            *self.raw_data["SecurityGroups"],
            *((i["InstanceId"], "routes_to") for i in self.raw_data["Instances"]),
        ]

    async def list_available_metrics(self) -> list[str]:
        return [
            "RequestCount",
            "HTTPCode_Backend_2XX",
            "HTTPCode_Backend_3XX",
            "HTTPCode_Backend_4XX",
            "HTTPCode_Backend_5XX",
            "HealthyHostCount",
            "UnHealthyHostCount",
            "Latency",
        ]

    async def get_metric(
        self,
        metric_name: str,
        time_range_start: AwareDatetime,
        time_range_end: AwareDatetime,
    ) -> list[Observation] | str:
        # Define appropriate statistics based on metric type
        if metric_name in [
            "RequestCount",
            "HTTPCode_Backend_2XX",
            "HTTPCode_Backend_3XX",
            "HTTPCode_Backend_4XX",
            "HTTPCode_Backend_5XX",
        ]:
            # Count-based metrics primarily use Sum
            statistics = ["Sum"]
        elif metric_name == "Latency":
            # Latency benefits from Average and percentiles
            statistics = ["Average", "p50", "p90", "p95", "p99"]
        elif metric_name in ["HealthyHostCount", "UnHealthyHostCount"]:
            # Host count metrics use Average, Min, and Max
            statistics = ["Average", "Minimum", "Maximum"]
        else:
            # Default statistics for other metrics
            statistics = ["Sum", "Average", "Minimum", "Maximum"]

        return await self._get_cloudwatch_metric_for_accessible_regions(
            Namespace="AWS/ELB",
            MetricName=metric_name,
            Dimensions=[
                {
                    "Name": "LoadBalancerName",
                    "Value": self.raw_data["LoadBalancerName"],
                }
            ],
            StartTime=time_range_start,
            EndTime=time_range_end,
            Period=300,
            Statistics=statistics,
        )



from typing import TYPE_CHECKING, Any

from fastmcp import FastMCP
from fastmcp.client.transports import FastMCPTransport, MCPConfigTransport
from fastmcp.mcp_config import MCPConfig
from fastmcp.utilities.mcp_config import mcp_config_to_servers_and_transports

if TYPE_CHECKING:
    from fastmcp.client import ClientTransport


class CompositeMCPTransport(MCPConfigTransport):
    """A transport that composes multiple MCP servers into a single transport.

    This is a slightly-modified version of the MCPConfigTransport, because the
    original implementation only prefixed tools if there was more than one MCP
    server configured.
    """

    def __init__(self, config: MCPConfig | dict, name_as_prefix: bool = True) -> None:
        self.config = config if isinstance(config, MCPConfig) else MCPConfig.from_dict(config)
        self._underlying_transports: list[ClientTransport] = []

        if not self.config.mcpServers:
            raise ValueError("No MCP servers defined in the config")

        name = FastMCP.generate_name("MCPRouter")
        self._composite_server = FastMCP[Any](name=name)

        for name, server, transport in mcp_config_to_servers_and_transports(self.config):
            self._underlying_transports.append(transport)
            self._composite_server.mount(server, prefix=name if name_as_prefix else None)

        self.transport = FastMCPTransport(mcp=self._composite_server)



from collections import defaultdict
from collections.abc import AsyncIterator
from datetime import UTC, datetime
from pathlib import Path
from typing import TypeVar, cast

import anyio
import networkx as nx
from anyio import Lock
from pydantic_core import from_json, to_json, to_jsonable_python

from unpage.utils import generate_contrasting_colors, print, strip_secrets

from .edges import Edge
from .nodes import NODE_REGISTRY, Node

_NodeType = TypeVar("_NodeType", bound="Node")


class Graph:
    _lock: Lock
    _identifier_mapping: defaultdict[str, set[str]]
    _loaded_at: datetime | None

    def __init__(self, path: Path | str | None = None) -> None:
        self._lock = Lock()
        self._identifier_mapping = defaultdict(set)
        self._path = Path(path) if path else None
        self._loaded_at = None

    @property
    def digraph(self) -> nx.DiGraph:
        # If the file has changed since the last load, force a reload.
        if (
            self._path
            and self._loaded_at
            and self._path.stat().st_mtime > self._loaded_at.timestamp()
        ):
            print("Graph file has changed since last load, forcing reload...")
            self._loaded_at = None

        if not self._loaded_at:
            if self._path:
                if not self._path.exists():
                    self._digraph = nx.DiGraph()
                    self._path.write_bytes(
                        to_json(
                            nx.node_link_data(
                                self._digraph,
                                edges="edges",
                            ),
                            indent=2,
                        )
                    )

                json = from_json(self._path.read_bytes())
                self._digraph = nx.node_link_graph(json, edges="edges")
            else:
                self._digraph = nx.DiGraph()

            # Convert the nodes to the proper Node objects.
            for _nid, node in self._digraph.nodes(data=True):
                node_data = node.pop("data")
                node_key = ":".join([node_data["node_source"], node_data["node_type"]])
                node["data"] = NODE_REGISTRY[node_key](
                    **node_data,
                    _graph=self,
                )

            self._loaded_at = datetime.now(UTC)

        return self._digraph

    async def add_node(self, node: Node) -> None:
        """Add a node to the graph."""
        async with self._lock:
            self.digraph.add_node(node.nid, data=node)

            for identifier in (node.nid, *await node.get_identifiers()):
                if not identifier:
                    continue
                self._identifier_mapping[identifier].add(node.nid)

    async def get_node(self, nid: str) -> Node:
        """Get a node from the graph."""
        try:
            node = self.digraph.nodes[nid]
        except KeyError as ex:
            raise LookupError(f"Node {nid} not found in graph") from ex
        return node["data"]

    async def get_node_safe(self, nid: str) -> Node | None:
        """Get a node from the graph. Returns None if no Node found with nid."""
        try:
            return await self.get_node(nid)
        except LookupError:
            return None

    async def iter_nodes(self) -> AsyncIterator[Node]:
        """Get all nodes in the graph."""
        for _, node in self.digraph.nodes(data=True):
            yield cast("Node", node["data"])

    async def iter_predecessors(
        self,
        node: Node,
    ) -> AsyncIterator[Node]:
        """Get all predecessors of a node."""
        for predecessor_nid in self.digraph.predecessors(node.nid):
            yield self.digraph.nodes[predecessor_nid]["data"]

    async def iter_successors(self, node: Node) -> AsyncIterator[Node]:
        """Get all successors of a node."""
        for successor_nid in self.digraph.successors(node.nid):
            yield self.digraph.nodes[successor_nid]["data"]

    async def iter_neighbors(self, node: Node) -> AsyncIterator[Node]:
        """Get all neighbors of a node."""
        async for neighbor in self.iter_predecessors(node):
            yield neighbor

        async for neighbor in self.iter_successors(node):
            yield neighbor

    async def iter_neighboring(
        self,
        node: Node,
        only_type: type[_NodeType] | tuple[type[_NodeType], ...],
    ) -> AsyncIterator[_NodeType]:
        """Get all neighbors of a node that are of a specific type."""
        only_types = only_type if isinstance(only_type, tuple) else (only_type,)

        async for neighbor in self.iter_neighbors(node):
            if not isinstance(neighbor, only_types):
                continue
            yield cast("_NodeType", neighbor)

    async def iter_neighborhood_edges(self, node: Node, max_depth: int = 1) -> AsyncIterator[Edge]:
        """Get all edges in the neighborhood of a node."""
        for reverse in (True, False):
            for source_nid, destination_nid in nx.bfs_edges(
                self._digraph, node.nid, depth_limit=max_depth, reverse=reverse
            ):
                if reverse:
                    source_node = self.digraph.nodes[destination_nid]["data"]
                    destination_node = self.digraph.nodes[source_nid]["data"]
                else:
                    source_node = self.digraph.nodes[source_nid]["data"]
                    destination_node = self.digraph.nodes[destination_nid]["data"]

                properties = (
                    self.digraph.edges[source_nid, destination_nid]
                    if not reverse
                    else self.digraph.edges[destination_nid, source_nid]
                )

                yield Edge(
                    source_node=source_node,
                    destination_node=destination_node,
                    properties=properties,
                )

    async def add_edge(self, edge: Edge) -> None:
        """Add an edge between nodes to the graph."""
        async with self._lock:
            self.digraph.add_node(edge.source_node.nid, data=edge.source_node)
            self.digraph.add_node(edge.destination_node.nid, data=edge.destination_node)

            edge_properties = edge.properties
            if "relationship_type" in edge_properties:
                edge_properties["label"] = edge_properties["relationship_type"]

            self.digraph.add_edge(
                edge.source_node.nid,
                edge.destination_node.nid,
                **edge_properties,
            )

    async def iter_edges(self) -> AsyncIterator[Edge]:
        """Get all edges in the graph."""
        for source_nid, destination_nid, properties in self.digraph.edges(data=True):
            source_node = self.digraph.nodes[source_nid]["data"]
            destination_node = self.digraph.nodes[destination_nid]["data"]

            yield Edge(
                source_node=source_node,
                destination_node=destination_node,
                properties=properties,
            )

    async def infer_edges(self) -> None:
        """Infer edges between nodes based on the identifier mapping."""
        edge_count = self.digraph.number_of_edges()
        limiter = anyio.CapacityLimiter(24)
        async with anyio.create_task_group() as tg:
            async for node in self.iter_nodes():
                tg.start_soon(self._infer_edges_for_node, node, limiter)
        print(f"Inferred {self.digraph.number_of_edges() - edge_count} edges")

    async def _infer_edges_for_node(self, node: Node, limiter: anyio.CapacityLimiter) -> None:
        async with limiter:
            for ref in await node.get_reference_identifiers():
                if isinstance(ref, tuple) and len(ref) == 2:
                    reference_identifier, relationship_type = ref
                else:
                    reference_identifier = ref
                    relationship_type = "related_to"

                if not reference_identifier:
                    continue

                related_node_ids = self._identifier_mapping.get(reference_identifier, set())

                if not related_node_ids:
                    continue

                if len(related_node_ids) > 1:
                    # TODO: Handle multiple related nodes (i.e. ambiguous edges).
                    continue

                related_node_id = next(iter(related_node_ids))
                related_node = await self.get_node(related_node_id)

                print(f"Inferred edge: {node.nid} --[ {relationship_type} ]--> {related_node.nid}")

                await self.add_edge(
                    Edge(
                        source_node=node,
                        destination_node=related_node,
                        properties={
                            "relationship_type": relationship_type,
                        },
                    )
                )

    async def get_topology(self) -> "Graph":
        topology = Graph()

        async for edge in self.iter_edges():
            source_node = edge.source_node.model_copy(update={"node_id": ""})
            destination_node = edge.destination_node.model_copy(update={"node_id": ""})

            # Add nodes (types) and edge between them
            await topology.add_edge(
                Edge(
                    source_node=source_node,
                    destination_node=destination_node,
                    properties={"label": edge.properties["relationship_type"]},
                )
            )

        return topology

    async def to_pydot(self) -> str:
        """Convert the graph to a Pydot graph."""
        node_types = {f"{node.node_source}:{node.node_type}" async for node in self.iter_nodes()}
        node_colors = dict(
            zip(node_types, generate_contrasting_colors(len(node_types)), strict=True)
        )

        graph: nx.DiGraph = nx.relabel_nodes(
            self._digraph,
            # The mapping can be a callable, despite the type hint.
            #
            # We also quote the labels, since identifiers can often contain
            # colons, which are special in Pydot.
            mapping=lambda old: f'"{old}"',  # type: ignore
            copy=True,
        )

        # Set the defaults.
        graph.graph["graph"] = {"nodesep": 1}
        graph.graph["node"] = {
            "shape": "box",
            "style": "rounded,filled",
            "fontname": "Helvetica",
            "margin": "0.25",
        }

        # Set the node colors.
        for _, node in graph.nodes(data=True):
            node_data = cast("Node", node["data"])
            node_key = f"{node_data.node_source}:{node_data.node_type}"
            fill_color, font_color = node_colors[node_key]
            node["fillcolor"] = fill_color
            node["fontcolor"] = font_color
            del node["data"]

        return nx.nx_pydot.to_pydot(graph).to_string()

    async def save(self, path: Path | str) -> None:
        """Save the graph to a file."""
        async with self._lock:
            data = nx.node_link_data(self._digraph, edges="edges")
            print("Scrubbing secrets from graph...")
            cleaned_data = strip_secrets(to_jsonable_python(data))
            path = Path(path)
            path.write_bytes(
                to_json(
                    {
                        **cleaned_data,
                        "_identifier_mapping": dict(self._identifier_mapping),
                    },
                    indent=2,
                )
            )



from typing import Any

from pydantic import BaseModel


class RootlyIncident(BaseModel):
    """Represents a Rootly incident."""

    id: str
    attributes: dict[str, Any]
    type: str = "incidents"


class RootlyIncidentPayload(BaseModel):
    """Represents a Rootly incident payload for webhooks."""

    incident: RootlyIncident



from typing import Any

import httpx


class RootlyClient:
    """Client for interacting with the Rootly API."""

    def __init__(self, api_key: str, base_url: str = "https://api.rootly.com/v1") -> None:
        self.api_key = api_key
        self.base_url = base_url
        self.headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/vnd.api+json",
            "Accept": "application/vnd.api+json",
        }

    async def _request(
        self, method: str, endpoint: str, json_data: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Make an HTTP request to the Rootly API."""
        url = f"{self.base_url}/{endpoint.lstrip('/')}"

        async with httpx.AsyncClient() as client:
            response = await client.request(
                method=method,
                url=url,
                headers=self.headers,
                json=json_data,
            )
            response.raise_for_status()
            return response.json()

    async def get_incident(self, incident_id: str) -> dict[str, Any]:
        """Get a specific incident by ID."""
        return await self._request("GET", f"/incidents/{incident_id}")

    async def list_incidents(self, params: dict[str, Any] | None = None) -> dict[str, Any]:
        """List incidents with optional filtering parameters."""
        if params:
            url = "/incidents?" + "&".join(f"{k}={v}" for k, v in params.items())
        else:
            url = "/incidents"
        return await self._request("GET", url)

    async def update_incident(self, incident_id: str, data: dict[str, Any]) -> dict[str, Any]:
        """Update an incident."""
        return await self._request("PUT", f"/incidents/{incident_id}", json_data=data)

    async def create_incident_event(
        self, incident_id: str, event_data: dict[str, Any]
    ) -> dict[str, Any]:
        """Create an event (like a status update) for an incident."""
        return await self._request(
            "POST", f"/incidents/{incident_id}/incident_events", json_data=event_data
        )

    async def get_incident_events(self, incident_id: str) -> dict[str, Any]:
        """Get events for an incident."""
        return await self._request("GET", f"/incidents/{incident_id}/events")

    async def create_incident_event_new(
        self, incident_id: str, event_data: dict[str, Any]
    ) -> dict[str, Any]:
        """Create a new event for an incident using the correct endpoint."""
        return await self._request("POST", f"/incidents/{incident_id}/events", json_data=event_data)

    async def mitigate_incident(self, incident_id: str) -> dict[str, Any]:
        """Mitigate an incident."""
        return await self._request("POST", f"/incidents/{incident_id}/mitigate")

    async def acknowledge_incident(self, incident_id: str) -> dict[str, Any]:
        """Acknowledge an incident."""
        return await self._request("POST", f"/incidents/{incident_id}/acknowledge")

    async def resolve_incident(self, incident_id: str) -> dict[str, Any]:
        """Resolve an incident."""
        return await self._request("POST", f"/incidents/{incident_id}/resolve")

    async def update_incident_event(self, event_id: str, data: dict[str, Any]) -> dict[str, Any]:
        """Update an incident event."""
        return await self._request("PUT", f"/events/{event_id}", json_data=data)

    async def delete_incident_event(self, event_id: str) -> dict[str, Any]:
        """Delete an incident event."""
        return await self._request("DELETE", f"/events/{event_id}")
